#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""Applications used as examples, for testing, or in simple situations"""

import _pickle
from numbers import Number
import pickle
import random
from typing import List, Optional
import requests
import logging
import time
import numpy as np

from dlg import droputils, drop_loaders
from dlg.apps.app_base import BarrierAppDROP
from dlg.apps.pyfunc import PyFuncApp
from dlg.data.drops.container import ContainerDROP
from dlg.data.drops import InMemoryDROP, FileDROP
from dlg.apps.branch import BranchAppDrop
from dlg.drop import track_current_drop
from dlg.meta import (
    dlg_float_param,
    dlg_string_param,
    dlg_bool_param,
    dlg_int_param,
    dlg_list_param,
    dlg_dict_param,
    dlg_component,
    dlg_batch_input,
    dlg_batch_output,
    dlg_streaming_input,
)
from dlg.exceptions import DaliugeException
from dlg.rpc import DropProxy

logger = logging.getLogger(f"dlg.{__name__}")


class NullBarrierApp(BarrierAppDROP):
    component_meta = dlg_component(
        "NullBarrierApp",
        "Null Barrier.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    """A BarrierAppDrop that doesn't perform any work"""

    def run(self):
        pass


##
# @brief DALiuGEApp
# @details A placeholder APP to aid construction of new applications.
# This is mainly useful (and used) when starting a new workflow from scratch.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag template
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.DALiuGEApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @par EAGLE_END
class DALiuGEApp(BarrierAppDROP):
    """A placeholder BarrierAppDrop that just aids the generation of the palette component"""


##
# @brief SleepApp
# @details A simple APP that sleeps the specified amount of time (0 by default).
# This is mainly useful (and used) to test graph translation and structure
# without executing real algorithms. Very useful for debugging.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param sleep_time 5/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/The number of seconds to sleep
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.SleepApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @par EAGLE_END


class SleepApp(BarrierAppDROP):
    """A BarrierAppDrop that sleeps the specified amount of time (0 by default)"""

    component_meta = dlg_component(
        "SleepApp",
        "Sleep App.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )
    sleep_time = dlg_float_param("sleep_time", 0)

    @track_current_drop
    def run(self):
        self._run()
        try:
            # If data is coming through a named port we load it from there.
            if isinstance(self.sleep_time, (InMemoryDROP, FileDROP, DropProxy)):
                logger.debug("Trying to read from %s", self.sleep_time)
                self.sleep_time = drop_loaders.load_pickle(self.sleep_time)
            time.sleep(self.sleep_time)
        except (TypeError, ValueError):
            logger.debug(
                "Found invalid sleep_time: %s. Resetting to 0. %s",
                self.sleep_time,
                type(self.sleep_time),
            )
            self.sleep_time = 0
            time.sleep(self.sleep_time)
        logger.info("%s slept for %s s", self.name, self.sleep_time)


##
# @brief CopyApp
# @details A simple APP that copies its inputs into its outputs.
# All inputs are copied into all outputs in the order they were declared in
# the graph. If an input is a container (e.g. a directory) it copies the
# content recursively.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param bufsize 65536/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/Buffer size
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.CopyApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
# @param io /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Input Output port
# @par EAGLE_END
class CopyApp(BarrierAppDROP):
    """
    A BarrierAppDrop that copies its inputs into its outputs.
    All inputs are copied into all outputs in the order they were declared in
    the graph.
    """

    bufsize = dlg_int_param("bufsize", 65536)

    component_meta = dlg_component(
        "CopyApp",
        "Copy App.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    def run(self):
        logger.debug("Using buffer size %d", self.bufsize)
        logger.info(
            "Copying data from inputs %s to outputs %s",
            [x.name for x in self.inputs],
            [x.name for x in self.outputs],
        )
        self.copyAll()
        logger.info(
            "Copy finished",
        )

    def copyAll(self):
        for inputDrop in self.inputs:
            self.copyRecursive(inputDrop)

        # logger.debug("Target checksum: %d", outputDrop.checksum)

    def copyRecursive(self, inputDrop):
        if isinstance(inputDrop, ContainerDROP):
            for child in inputDrop.children:
                self.copyRecursive(child)
        else:
            for outputDrop in self.outputs:
                droputils.copyDropContents(inputDrop, outputDrop, bufsize=self.bufsize)


##
# @brief SleepAndCopyApp
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param sleep_time 5/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/The number of seconds to sleep
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.SleepAndCopyApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @par EAGLE_END
class SleepAndCopyApp(SleepApp, CopyApp):
    """A combination of the SleepApp and the CopyApp. It sleeps, then copies"""

    def run(self):
        SleepApp.run(self)
        CopyApp.run(self)


##
# @brief RandomArrayApp
# @details A testing APP that does not take any input and produces a random array of
# type int64, if integer is set to True, else of type float64.
# size indicates the number of elements ranging between the values low and high.
# The resulting array will be send to all connected output apps.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param size 100/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/The size of the array
# @param low 0/Float/ApplicationArgument/NoPort/ReadWrite//False/False/Low value of range in array [inclusive]
# @param high 1/Float/ApplicationArgument/NoPort/ReadWrite//False/False/High value of range of array [exclusive]
# @param integer True/Boolean/ApplicationArgument/NoPort/ReadWrite//False/False/Generate integer array?
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.RandomArrayApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param array /Object.Array/ApplicationArgument/OutputPort/ReadWrite//False/False/random array
# @par EAGLE_END
class RandomArrayApp(BarrierAppDROP):
    """
    A BarrierAppDrop that generates an array of random numbers. It does
    not require any inputs and writes the generated array to all of its
    outputs.

    Keywords:

    integer:  bool [True], generate integer array
    low:      float, lower boundary (will be converted to int for integer arrays)
    high:     float, upper boundary (will be converted to int for integer arrays)
    size:     int, number of array elements
    """

    component_meta = dlg_component(
        "RandomArrayApp",
        "Random Array App.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    # default values
    integer = dlg_bool_param("integer", True)
    low = dlg_float_param("low", 0)
    high = dlg_float_param("high", 100)
    size = dlg_int_param("size", 100)
    marray = []

    def initialize(self, keep_array=False, **kwargs):
        super(RandomArrayApp, self).initialize(**kwargs)
        self._keep_array = keep_array


    @track_current_drop
    def run(self):
        self._run()
        # At least one output should have been added
        outs = self.outputs
        if len(outs) < 1:
            raise Exception("At least one output should have been added to %r" % self)
        logger.info("Generating %d random numbers between %f and %f", self.size, self.low, self.high)
        marray = self.generateRandomArray()
        if self._keep_array:
            self.marray = marray
        for o in outs:
            d = pickle.dumps(marray)
            o.len = len(d)
            o.write(d)

    def generateRandomArray(self):
        if self.integer:
            # generate an array of self.size integers with numbers between
            # slef.low and self.high
            marray = np.random.randint(int(self.low), int(self.high), size=(self.size))
        else:
            # generate an array of self.size floats with numbers between
            # self.low and self.high
            marray = (np.random.random(size=self.size) + self.low) * self.high
        return marray

    def _getArray(self):
        return self.marray


##
# @brief AverageArrays
# @details A testing APP that takes multiple numpy arrays on input and calculates
# the mean or the median, depending on the value provided in the method parameter.
# Users can add as many producers to the input array port as required and the resulting array
# will also be send to all connected output apps.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param method mean/Select/ApplicationArgument/NoPort/ReadWrite/mean,median/False/False/The method used for averaging
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.AverageArraysApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param array /Object.Array/ApplicationArgument/InputOutput/ReadWrite//False/False/Port for the array(s)
# @par EAGLE_END
class AverageArraysApp(BarrierAppDROP):
    """
    A BarrierAppDrop that averages arrays received on input. It requires
    multiple inputs and writes the generated average vector to all of its
    outputs.
    The input arrays are assumed to have the same number of elements and
    the output array will also have that same number of elements.

    Keywords:

    method:  string <['mean']|'median'>, use mean or median as method.
    """

    from numpy import mean, median

    component_meta = dlg_component(
        "AverageArraysApp",
        "Average Array App.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    # default values
    methods = ["mean", "median"]
    method = dlg_string_param("method", methods[0])

    def __init__(self, oid, uid, **kwargs):
        super().__init__(oid, kwargs)
        self.marray = []

    def run(self):
        # At least one output should have been added

        outs = self.outputs
        if len(outs) < 1:
            raise Exception("At least one output should have been added to %r" % self)
        self.getInputArrays()
        self._avg = self.averageArray()
        for o in outs:
            d = pickle.dumps(self._avg)
            o.len = len(d)
            o.write(d)  # average across inputs

    def getInputArrays(self):
        """
        Create the input array from all inputs received. Shape is
        (<#inputs>, <#elements>), where #elements is the length of the
        vector received from one input.
        """
        ins = self.inputs
        if len(ins) < 1:
            raise Exception("At least one input should have been added to %r" % self)
        marray = []
        for inp in ins:
            sarray = droputils.allDropContents(inp)
            if len(sarray) == 0:
                print("Input does not contain data!")
            else:
                sarray = pickle.loads(sarray)
                if isinstance(sarray, (list, tuple, np.ndarray)):
                    marray.extend(list(sarray))
                else:
                    marray.append(sarray)
        self.marray = marray

    def averageArray(self):
        method_to_call = getattr(np, self.method)
        return method_to_call(self.marray, axis=0)


##
# @brief GenericGatherApp
# @details App that reads all its inputs and simply writes them in
# concatenated to all its outputs. This can be used stand-alone or
# as part of a Gather. It does not do anything to the data, just
# passing it on.
#
# @par EAGLE_START
# @param construct Gather
# @param category DALiuGEApp
# @param tag daliuge
# @param num_of_inputs 4/Integer/ConstructParameter/NoPort/ReadWrite//False/False/The Gather width, stating how many inputs each Gather instance will handle
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.GenericGatherApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input /Object/ApplicationArgument/InputPort/ReadWrite//False/False/0-base placeholder port for inputs
# @param output /Object/ApplicationArgument/OutputPort/ReadWrite//False/False/Placeholder port for outputs
# @par EAGLE_END
class GenericGatherApp(BarrierAppDROP):
    component_meta = dlg_component(
        "GenericGatherApp",
        "Gather multiple inputs",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    # automatically populated by scatter node
    num_of_inputs: int = dlg_int_param("num_of_inputs", 1)

    def readWriteData(self):
        inputs = self.inputs
        outputs = self.outputs
        for output in outputs:
            for ipt in inputs:
                value = droputils.allDropContents(ipt)
                output.write(value)

    def run(self):
        self.readWriteData()


##
# @brief DictGatherApp
# @details App packs all data on input into a dictionary using the input drop's names as keys and the reading the
# dict values from the input drops. This app can be used stand-alone without a gather construct.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param value_dict value_dict/Jason/ApplicationArgument/NoPort/ReadWrite//False/False/The value dictionary can be initialized here
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.DictGatherApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input /Object/ApplicationArgument/InputPort/ReadWrite//False/False/0-base placeholder port for inputs
# @param output /Object/ApplicationArgument/OutputPort/ReadWrite//False/False/Placeholder port for outputs
# @par EAGLE_END
class DictGatherApp(BarrierAppDROP):
    component_meta = dlg_component(
        "DictGatherApp",
        "Collect multiple inputs into a dictionary",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )
    value_dict = dlg_dict_param("value_dict", {})

    def initialize(self, **kwargs):
        super(DictGatherApp, self).initialize(**kwargs)
        self.kwargs = kwargs

    def readWriteData(self):
        inputs = self.inputs
        outputs = self.outputs
        # for input in inputs:
        #     total_len += input.size
        # logger.debug(f">>>> writing {inputs} to {outputs}")
        for output in outputs:
            for ipt in inputs:
                value = droputils.allDropContents(ipt)
                self.value_dict[ipt.name] = pickle.loads(value)
                for aa_key, aa_dict in self.kwargs["applicationArgs"].items():
                    if aa_key not in self.value_dict and aa_dict["value"]:
                        self.value_dict[aa_key] = aa_dict["value"]
            logger.debug(
                "Writing %s to %s",
                self.value_dict,
                output.name,
            )
            output.write(pickle.dumps(self.value_dict))

            # logger.debug(f">>> written {d} to {output}")


##
# @brief ArrayGatherApp
# @details App appends all input daata to a list. This app can be used stand-alone without a gather construct.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param value_array value_array/array/ApplicationArgument/OutputPort/ReadWrite//False/False/The value array can be initialized here
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.ArrayGatherApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input /Object/ComponentParameter/InputPort/ReadWrite//False/False/0-base placeholder port for inputs
# @par EAGLE_END
class ArrayGatherApp(BarrierAppDROP):
    component_meta = dlg_component(
        "ArrayGatherApp",
        "Collect multiple inputs into an array",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )
    # value_list = dlg_list_param("value_list", [])

    def initialize(self, **kwargs):
        super(ArrayGatherApp, self).initialize(**kwargs)
        self.kwargs = kwargs

    def readWriteData(self):
        inputs = self.inputs
        outputs = self.outputs
        for output in outputs:
            for ipt in inputs:
                value = droputils.allDropContents(ipt)
                self.value_list.append(pickle.loads(value))
            output.write(pickle.dumps(self.value_list))

    def run(self):
        self.value_list = []
        self.readWriteData()


##
# @brief GenericNpyGatherApp
# @details A BarrierAppDrop that combines one or more inputs using cumulative operations.
# @par EAGLE_START
# @param category DALiuGEApp
# @param construct Gather
# @param tag daliuge
# @param num_of_inputs 4/Integer/ConstructParameter/NoPort/ReadWrite//False/False/The Gather width, stating how many inputs each Gather instance will handle
# @param function sum/Select/ApplicationArgument/NoPort/ReadWrite/sum,prod,min,max,add,multiply,maximum,minimum/False/False/The function used for gathering
# @param reduce_axes None/String/ApplicationArgument/NoPort/ReadOnly//False/False/The array axes to reduce, None reduces all axes for sum, prod, max, min functions
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.GenericNpyGatherApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param array_in /Object.Array/ApplicationArgument/InputPort/ReadWrite//False/False/Port for the input array(s)
# @param array_out /Object.Array/ApplicationArgument/OutputPort/ReadWrite//False/False/reduced array
# @par EAGLE_END
class GenericNpyGatherApp(BarrierAppDROP):
    """
    A BarrierAppDrop that reduces then gathers one or more inputs using cumulative operations.
    function:  string <'sum'|'prod'|'min'|'max'|'add'|'multiply'|'maximum'|'minimum'>.

    """

    component_meta = dlg_component(
        "GenericNpyGatherApp",
        "Generic Npy Gather App.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    # reduce and combine operation pair names
    # reduce operation reduces the dimensionality of a ndarray
    # gather operation combines ndarrays and retains dimensionality
    functions = {
        # reduce and gather (output dimension is reduced)
        "sum": "add",  # sum reduction of inputs along an axis first then gathers across drops
        "prod": "multiply",  # prod reduction of inputs along an axis first then gathers across drops
        "max": "maximum",  # max reduction of input along an axis first then gathers across drops
        "min": "minimum",  # min reduction of input along an axis first then gathers across drops
        # gather only
        "add": None,  # elementwise addition of inputs, ndarrays must be of same shape
        "multiply": None,  # elementwise multiplication of inputs, ndarrays must be of same shape
        "maximum": None,  # elementwise maximums of inputs, ndarrays must be of same shape
        "minimum": None,  # elementwise minimums of inputs, ndarrays must be of same shape
    }
    function: str = dlg_string_param("function", "sum")  # type: ignore
    reduce_axes: list = dlg_list_param("reduce_axes", "None")  # type: ignore

    def run(self):
        if len(self.inputs) < 1:
            raise Exception(f"At least one input should have been added to {self}")
        if len(self.outputs) < 1:
            raise Exception(f"At least one output should have been added to {self}")
        if self.function not in self.functions:
            raise Exception(f"Function {self.function} not supported by {self}")

        result = (
            self.reduce_gather_inputs()
            if self.functions[self.function] is not None
            else self.gather_inputs()
        )

        for o in self.outputs:
            drop_loaders.save_numpy(o, result)

    def reduce_gather_inputs(self):
        """reduces then gathers each input drop interpreted as an npy drop"""
        result: Optional[Number] = None
        reduce = getattr(np, f"{self.function}")
        gather = getattr(np, f"{self.functions[self.function]}")
        for ipt in self.inputs:
            data = drop_loaders.load_numpy(ipt)
            # skip gather for the first input
            result = (
                reduce(data, axis=self.reduce_axes)
                if result is None
                else gather(
                    result,
                    reduce(data, axis=self.reduce_axes),
                )
            )
        return result

    def gather_inputs(self):
        """gathers each input drop interpreted as an npy drop"""
        result: Optional[Number] = None
        gather = getattr(np, f"{self.function}")
        for ipt in self.inputs:
            data = drop_loaders.load_numpy(ipt)
            # assign instead of gather for the first input
            # result = data if result is None else gather(result, data, allow_pickle=True)
            result = data if result is None else gather(result, data)
        return result


##
# @brief HelloWorldApp
# @details A simple APP that implements the standard Hello World in DALiuGE.
# It allows to change 'World' with some other string and it also permits
# to connect the single output port to multiple sinks, which will all receive
# the same message. App does not require any input.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param greet World/String/ApplicationArgument/InputPort/ReadWrite//False/False/What appears after 'Hello '
# @param hello "world"/Object/ApplicationArgument/OutputPort/ReadWrite//False/False/message
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.HelloWorldApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @par EAGLE_END
class HelloWorldApp(BarrierAppDROP):
    """
    An App that writes 'Hello World!' or 'Hello <greet>!' to all of
    its outputs.

    Keywords:
    greet:   string, [World], whom to greet.
    """

    component_meta = dlg_component(
        "HelloWorldApp",
        "Hello World App.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    greet = dlg_string_param("greet", "World")

    def run(self):
        ins = self.inputs
        # if no inputs use the parameter else use the input
        if len(ins) == 0:
            self.greeting = "Hello %s" % self.greet
        elif len(ins) != 1:
            raise Exception("Only one input expected for %r" % self)
        else:  # the input is expected to be a vector. We'll use the first element
            try:
                phrase = str(pickle.loads(droputils.allDropContents(ins[0]))[0])
            except _pickle.UnpicklingError:
                phrase = str(droputils.allDropContents(ins[0]), encoding="utf-8")
            self.greeting = f"Hello {phrase}"
        logger.debug("Greeting is %s", self.greeting)

        outs = self.outputs
        if len(outs) < 1:
            raise Exception("At least one output should have been added to %r" % self)
        for o in outs:
            o.len = len(self.greeting.encode())
            o.write(self.greeting.encode())  # greet across all outputs


##
# @brief UrlRetrieveApp
# @details A simple APP that retrieves the content of a URL and writes
# it to all outputs.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param url None/String/ApplicationArgument/NoPort/ReadWrite//False/False/The URL to retrieve
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.UrlRetrieveApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param content /String/ApplicationArgument/OutputPort/ReadWrite//False/False/content read from URL
# @par EAGLE_END
class UrlRetrieveApp(BarrierAppDROP):
    """
    An App that retrieves the content of a URL

    Keywords:
    URL:   string, URL to retrieve.
    """

    component_meta = dlg_component(
        "UrlRetrieveApp",
        "URL Retrieve App",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    url = dlg_string_param("url", "")

    def run(self):
        try:
            logger.info("Accessing URL %s", self.url)
            u = requests.get(self.url, timeout=30)
        except requests.exceptions.RequestException as e:
            raise e.reason

        outs = self.outputs
        if len(outs) < 1:
            raise Exception("At least one output should have been added to %r" % self)
        for o in outs:
            o.len = len(u.content)
            o.write(u.content)  # send content to all outputs


##
# @brief GenericScatterApp
# @details An APP that splits about any object that can be converted to a numpy array
# into as many parts as the app has outputs, provided that the initially converted numpy
# array has enough elements. The return will be a numpy array of arrays, where the first
# axis is of length len(outputs). The modulo remainder of the length of the original array and
# the number of outputs will be distributed across the first len(outputs)-1 elements of the
# resulting array.
# @par EAGLE_START
# @param category DALiuGEApp
# @param construct Scatter
# @param tag daliuge
# @param num_of_copies 4/Integer/ConstructParameter/NoPort/ReadWrite//False/False/Specifies the number of replications of the content of the scatter construct
# @param array_in /Object.Array/ApplicationArgument/InputPort/ReadWrite//False/False/A numpy array of arrays, where the first axis is of length <numSplit>
# @param object_out /Object/ApplicationArgument/OutputPort/ReadWrite//False/False/reduced array or single element, depending on element flag.
# @param element False/Boolean/ApplicationArgument/NoPort/ReadWrite//False/False/if True the outputs of each of the splits will be the first element of the split array, rather than the split array.
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.GenericScatterApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @par EAGLE_END
class GenericScatterApp(BarrierAppDROP):
    """
    An APP that splits an object that has a len attribute into <numSplit> parts and
    returns a numpy array of arrays, where the first axis is of length <numSplit>.
    """

    component_meta = dlg_component(
        "GenericScatterApp",
        "Scatter an array like object into numSplit parts",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    # automatically populated by scatter node
    num_of_copies: int = dlg_int_param("num_of_copies", 1)
    element: bool = dlg_bool_param("element", False)

    @track_current_drop
    def run(self):
        numSplit = self.num_of_copies
        cont = droputils.allDropContents(self.inputs[0])
        # if the data is of type string it is not pickled, but stored as a binary string.
        try:
            inpArray = pickle.loads(cont)
        except pickle.PickleError:
            inpArray = cont.decode()
        nObj = np.array(inpArray)
        result = np.array_split(nObj, numSplit)
        for i in range(numSplit):
            o = self.outputs[i]
            if not self.element:
                d = pickle.dumps(result[i])
            else:
                d = pickle.dumps(result[i][0])
            o.len = len(d)
            o.write(d)  # average across inputs


##
# @brief GenericNpyScatterApp
# @details An APP that splits about any axis on any npy format data drop
# into as many part./run    s as the app has outputs, provided that the initially converted numpy
# array has enough elements. The return will be a numpy array of arrays, where the first
# axis is of length len(outputs). The modulo remainder of the length of the original array and
# the number of outputs will be distributed across the first len(outputs)-1 elements of the
# resulting array.
# @par EAGLE_START
# @param construct Scatter
# @param category DALiuGEApp
# @param tag daliuge
# @param num_of_copies 4/Integer/ConstructParameter/NoPort/ReadWrite//False/False/Specifies the number of replications of the content of the scatter construct
# @param scatter_axes /String/ApplicationArgument/NoPort/ReadWrite//False/False/The axes to split input ndarrays on, e.g. [0,0,0], length must match the number of input ports
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.GenericScatterApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param dropclass dlg.apps.simple.GenericNpyScatterApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param array_in /Object.Array/ApplicationArgument/InputPort/ReadWrite//False/False/A numpy array of arrays
# @param array_out /Object.Array/ApplicationArgument/OutputPort/ReadWrite//False/False/reduced array
# @par EAGLE_END
class GenericNpyScatterApp(BarrierAppDROP):
    """
    An APP that splits an object that has a len attribute into <num_of_copies> parts and
    returns a numpy array of arrays.
    """

    component_meta = dlg_component(
        "GenericNpyScatterApp",
        "Scatter an array like object into <num_of_copies> parts",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    # automatically populated by scatter node
    num_of_copies: int = dlg_int_param("num_of_copies", 1)
    scatter_axes: List[int] = dlg_list_param("scatter_axes", "[0]")

    def run(self):
        if len(self.inputs) * self.num_of_copies != len(self.outputs):
            raise DaliugeException(
                f"expected {len(self.inputs) * self.num_of_copies} outputs,\
                 got {len(self.outputs)}"
            )
        if len(self.inputs) != len(self.scatter_axes):
            raise DaliugeException(
                f"expected {len(self.inputs)} axes,\
                 got {len(self.scatter_axes)}, {self.scatter_axes}"
            )

        # split it as many times as we have outputs
        self.num_of_copies = self.num_of_copies

        for in_index in range(len(self.inputs)):
            nObj = drop_loaders.load_numpy(self.inputs[in_index])
            try:
                result = np.array_split(
                    nObj, self.num_of_copies, axis=self.scatter_axes[in_index]
                )
            except IndexError as err:
                raise err
            for split_index in range(self.num_of_copies):
                out_index = in_index * self.num_of_copies + split_index
                drop_loaders.save_numpy(self.outputs[out_index], result[split_index])


class SimpleBranch(BranchAppDrop, NullBarrierApp):
    """
    Simple branch app that is told the result of its condition.
    We are keeping this not to break existing graphs.
    """

    def initialize(self, **kwargs):
        self.result = self._popArg(kwargs, "result", True)
        BranchAppDrop.initialize(self, **kwargs)

    def run(self):
        pass

    def condition(self):
        return self.result


##
# @brief Branch
# @details A branch application that copies the input to either the 'true' or the 'false' output depending on the result of
# the provided conditional function. The conditional function can be specified either in-line or as an external function and has
# to return a boolean value.
# The inputs of the application are passed on as arguments to the conditional function. The conditional function needs to return
# a boolean value, but the application will copy the input data to the true or false output, depending on the result of the
# conditional function.
# @par EAGLE_START
# @param category Branch
# @param tag daliuge
# @param func_name condition/String/ComponentParameter/NoPort/ReadWrite//False/False/Python conditional function name. This can also be a valid import path to an importable function.
# @param func_code def condition(x): return (x > 0)/String/ComponentParameter/NoPort/ReadWrite//False/False/Python function code for the branch condition. Modify as required. Note that func_name above needs to match the defined name here.
# @param x /Object/ComponentParameter/InputPort/ReadWrite//False/False/Port carrying the input which is also used in the condition function. Note that the name of the parameter has to match the argument of the condition function.
# @param true  /Object/ComponentParameter/OutputPort/ReadWrite//False/False/If condition is true the input will be copied to this port
# @param false /Object/ComponentParameter/OutputPort/ReadWrite//False/False/If condition is false the input will be copied to this port
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.Branch/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @par EAGLE_END
class Branch(PyFuncApp):
    """
    A branch application that copies the input to either the 'true' or the 'false' output depending on the result of
    the provided conditional function. The conditional function can be specified either in-line or as an external function and has
    to return a boolean value.
    The inputs of the application are passed on as arguments to the conditional function. The conditional function needs to return
    a boolean value, but the application will copy the input data to the true or false output, depending on the result of the
    conditional function.
    """

    bufsize = dlg_int_param("bufsize", 65536)
    result = dlg_bool_param("result", False)

    def write_results(self,result:bool=False):
        """
        Copy the input to the output identified by the condition function.
        """
        if result and isinstance(result, bool):
            self.result = result
        if not self.outputs:
            return

        go_result = str(self.result).lower()
        nogo_result = str(not self.result).lower()

        try:
            nogo_drop = getattr(self, nogo_result)
        except AttributeError:
            logger.error("There is no Drop associated with the False condition; "
                         "a runtime failure has occured.")
            self.setError()
            return
        try:
            go_drop = getattr(self, go_result)
        except AttributeError:
            logger.error("There is no Drop associated with the True condition; "
                         "a runtime failure has occured.")
            self.setError()
            return

        logger.info("Sending skip to port: %s: %s", str(nogo_result), getattr(self,nogo_result))
        nogo_drop.skip()  # send skip to correct branch

        if self.inputs and hasattr(go_drop, "write"):
            droputils.copyDropContents(  # send data to correct branch
                self.inputs[0], go_drop, bufsize=self.bufsize
            )
        else:  # this enables a branch based only on the condition function
            d = pickle.dumps(self.parameters[self.argnames[0]])
            # d = self.parameters[self.argnames[0]]
            if hasattr(go_drop, "write"):
                go_drop.write(d)


##
# @brief PickOne
# @details App that picks the first element of an input list, passes that
# to all outputs, except the first one. The first output is used to pass
# the remaining array on. This app is useful for a loop.
#
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.GenericScatterApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param dropclass dlg.apps.simple.PickOne/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param rest_array /Object.Array/ApplicationArgument/InputOutput/ReadWrite//False/False/List of elements
# @param element /Object.element/ApplicationArgument/OutputPort/ReadWrite//False/False/first element
# @par EAGLE_END
class PickOne(BarrierAppDROP):
    """
    Simple app picking one element at a time. Good for Loops.
    """

    def initialize(self, **kwargs):
        BarrierAppDROP.initialize(self, **kwargs)

    def readData(self):
        ipt = self.inputs[0]
        data = pickle.loads(droputils.allDropContents(ipt))
        # data = droputils.allDropContents(input)
        # data = dill.loads(base64.b64decode(data))

        # make sure we always have a ndarray with at least 1dim.
        if type(data) not in (list, tuple) and not isinstance(data, (np.ndarray)):
            logger.warning("Data type not in [list, tuple]: %s", data)
            raise TypeError
        if isinstance(data, np.ndarray) and data.ndim == 0:
            data = np.array([data])
        else:
            data = np.array(data)
        value = data[0] if len(data) else None
        rest = data[1:] if len(data) >= 1 else np.array([])
        return value, rest

    def writeData(self, value, rest):
        """
        Prepare the data and write to all outputs
        """
        # write rest to array output
        # and value to every other output
        for output in self.outputs:
            if output.name == "rest_array":
                d = pickle.dumps(rest)
                output.len = len(d)
            else:
                d = pickle.dumps(value)
                output.len = len(d)
            output.write(d)

    def run(self):
        value, rest = self.readData()
        self.writeData(value, rest)


##
# @brief ListAppendThrashingApp
# @details A testing APP that appends a random integer to a list num times.
# This is a CPU intensive operation and can thus be used to provide a test for application threading
# since this operation will not yield.
# The resulting array will be sent to all connected output apps.
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag test
# @param size 100/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/the size of the array
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.simple.GenericScatterApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param dropclass dlg.apps.simple.ListAppendThrashingApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name simple/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param array /Object.Array/ApplicationArgument/OutputPort/ReadWrite//False/False/random array
# @par EAGLE_END
class ListAppendThrashingApp(BarrierAppDROP):
    """
    A BarrierAppDrop that appends random integers to a list N times. It does
    not require any inputs and writes the generated array to all of its
    outputs.

    Keywords:

    size:     int, number of array elements
    """

    compontent_meta = dlg_component(
        "ListAppendThrashingApp",
        "List Append Thrashing",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    def initialize(self, **kwargs):
        self.size = self._popArg(kwargs, "size", 100)
        self.marray = []
        super(ListAppendThrashingApp, self).initialize(**kwargs)

    def run(self):
        # At least one output should have been added
        outs = self.outputs
        if len(outs) < 1:
            raise Exception("At least one output should have been added to %r" % self)
        self.marray = self.generateArray()
        for o in outs:
            d = pickle.dumps(self.marray)
            o.len = len(d)
            o.write(pickle.dumps(self.marray))

    def generateArray(self):
        # This operation is wasteful to simulate an N^2 operation.
        marray = []
        for _ in range(int(self.size)):
            marray = []
            for _ in range(int(self.size)):
                marray.append(random.random())
        return marray

    def _getArray(self):
        return self.marray



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import logging
import os
import shutil

from dlg.data.drops.file import FileDROP
from dlg.ddap_protocol import DROPRel, DROPLinkType
from dlg.data.drops.data_base import PathBasedDrop
from dlg.data.drops.container import ContainerDROP
from dlg.exceptions import InvalidDropException, InvalidRelationshipException
from dlg.meta import dlg_bool_param

logger = logging.getLogger(f"dlg.{__name__}")


# TODO: This needs some more work
##
# @brief Directory
# @details A ContainerDROP that represents a filesystem directory. It only allows
# FileDROPs and DirectoryContainers to be added as children. Children
# can only be added if they are placed directly within the directory
# represented by this DirectoryContainer.
# @par EAGLE_START
# @param category Directory
# @param tag future
# @param dropclass dlg.data.drops.directorycontainer.DirectoryContainer/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param base_name directorycontainer/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param check_exists True/Boolean/ApplicationArgument/NoPort/ReadWrite//False/False/Perform a check to make sure the file path exists before proceeding with the application
# @param dirname /String/ApplicationArgument/NoPort/ReadWrite//False/False/"Directory name/path"
# @param block_skip False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/If set the drop will block a skipping chain until the last producer has finished and is not also skipped.
# @param io /Object/ApplicationArgument/OutputPort/ReadWrite//False/False/Input Output port
# @par EAGLE_END
class DirectoryContainer(PathBasedDrop, ContainerDROP):
    """
    A ContainerDROP that represents a filesystem directory. It only allows
    FileDROPs and DirectoryContainers to be added as children. Children
    can only be added if they are placed directly within the directory
    represented by this DirectoryContainer.
    """

    check_exists = dlg_bool_param("check_exists", True)

    def initialize(self, **kwargs):
        ContainerDROP.initialize(self, **kwargs)

        if "dirname" not in kwargs:
            raise InvalidDropException(
                self, 'DirectoryContainer needs a "dirname" parameter'
            )

        directory = kwargs["dirname"]

        logger.debug("Checking existence of %s %s", directory, self.check_exists)
        if "check_exists" in kwargs and kwargs["check_exists"] is True:
            if not os.path.isdir(directory):
                raise InvalidDropException(self, "%s is not a directory" % (directory))

        self._path = self.get_dir(directory)

    def addChild(self, child):
        if isinstance(child, (FileDROP, DirectoryContainer)):
            path = child.path
            if os.path.dirname(path) != self.path:
                raise InvalidRelationshipException(
                    DROPRel(child, DROPLinkType.CHILD, self),
                    "Child DROP is not under %s" % (self.path),
                )
            ContainerDROP.addChild(self, child)
        else:
            raise TypeError("Child DROP is not of type FileDROP or DirectoryContainer")

    def delete(self):
        shutil.rmtree(self._path)

    def exists(self):
        return os.path.isdir(self._path)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2019
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import json
import logging
import re
import subprocess
import time

logger = logging.getLogger(f"dlg.{__name__}")


class ListTokens(object):
    STRING, COMMA, RANGE_SEP, MULTICASE_START, MULTICASE_END = range(5)


def _list_tokenizer(s):
    buff = []
    in_square_brackets = False
    for char in s:
        if char == "-" and in_square_brackets:
            yield ListTokens.STRING, "".join(buff)
            buff = []
            yield ListTokens.RANGE_SEP, None
        elif char == ",":
            if buff:
                yield ListTokens.STRING, "".join(buff)
                buff = []
            yield ListTokens.COMMA, None
        elif char == "[":
            in_square_brackets = True
            if buff:
                yield ListTokens.STRING, "".join(buff)
                buff = []
            yield ListTokens.MULTICASE_START, None
        elif char == "]":
            in_square_brackets = False
            if buff:
                yield ListTokens.STRING, "".join(buff)
                buff = []
            yield ListTokens.MULTICASE_END, None
        else:
            buff.append(char)
    if buff:
        yield ListTokens.STRING, "".join(buff)
        buff = []


def _parse_list_tokens(token_iter):
    def finish_element(sub_values, range_start):
        if sub_values:
            values.extend(sub_values)
        elif range_start is not None:
            range_end = values.pop()
            str_len = max(len(range_start), len(range_end))
            str_format = "%%0%dd" % str_len
            num_vals = [
                str_format % num for num in range(int(range_start), int(range_end) + 1)
            ]
            values.extend(num_vals)

    values = []
    sub_values = []
    range_start = None
    while True:
        try:
            token, value = next(token_iter)
        except StopIteration:
            finish_element(sub_values, range_start)
            return values
        if token == ListTokens.MULTICASE_END:
            finish_element(sub_values, range_start)
            return values
        if token == ListTokens.MULTICASE_START:
            prefix = ""
            if values:
                prefix = values.pop()
            sub_values = _parse_list_tokens(token_iter)
            sub_values = [prefix + s for s in sub_values]
        if token == ListTokens.RANGE_SEP:
            range_start = values.pop()
        elif token == ListTokens.COMMA:
            finish_element(sub_values, range_start)
            sub_values = None
            range_start = None
        elif token == ListTokens.STRING:
            if sub_values:
                sub_values = [s + value for s in sub_values]
            else:
                values.append(value)


def list_as_string(s):
    """'a008,b[072-073,076]' --> ['a008', 'b072', 'b073', 'b076']"""
    return _parse_list_tokens(iter(_list_tokenizer(s)))


def find_numislands(physical_graph_template_file):
    """
    Given the physical graph data extract the graph name and the total number of
    nodes. We are not making a decision whether the island managers are running
    on separate nodes here, thus the number is the sum of all island
    managers and node managers. The values are only populated if not given on the
    init already.
    TODO: We will probably need to do the same with job duration and CPU number
    """
    if not physical_graph_template_file:
        return None, None, physical_graph_template_file

    with open(physical_graph_template_file, "r") as f:
        pgt_data = json.load(f, strict=False)
    try:
        (pgt_name, pgt) = pgt_data
    except TypeError as e:
        raise ValueError(type(pgt_data)) from e
    try:
        nodes = list(map(lambda x: x["node"], pgt))
        islands = list(map(lambda x: x["island"], pgt))
    except KeyError:
        return None, None, pgt_name
    num_islands = len(dict(zip(islands, range(len(islands)))))
    num_nodes = len(dict(zip(nodes, range(len(nodes)))))
    return num_islands, num_nodes, pgt_name


def label_job_dur(job_dur):
    """
    e.g. 135 min --> 02:15:00
    """
    seconds = job_dur * 60
    minute, sec = divmod(seconds, 60)
    hour, minute = divmod(minute, 60)
    return "%02d:%02d:%02d" % (hour, minute, sec)


def num_daliuge_nodes(num_nodes: int, run_proxy: bool):
    """
    Returns the number of daliuge nodes available to run workflow
    """
    if run_proxy:
        ret = num_nodes - 1  # exclude the proxy node
    else:
        ret = num_nodes - 0  # exclude the data island node?
    if ret <= 0:
        raise Exception("Not enough nodes {0} to run DALiuGE.".format(num_nodes))
    return ret


def find_node_ips():
    query = subprocess.check_output(
        [
            r"kubectl get nodes --selector=kubernetes.io/role!=master -o jsonpath={.items[*].status.addresses[?\(@.type==\"InternalIP\"\)].address}"
        ],
        shell=True,
    )
    node_ips = query.decode(encoding="utf-8").split(" ")
    return node_ips


def find_service_ips(num_expected, retries=3, timeout=10):
    pattern = r"^daliuge-daemon-service-.*\s*ClusterIP\s*\d+\.\d+\.\d+\.\d+"
    ip_pattern = r"\d+\.\d+\.\d+\.\d+"
    ips = []
    attempts = 0
    while len(ips) < num_expected and attempts < retries:
        ips = []
        query = subprocess.check_output(
            [r"kubectl get svc -o wide"], shell=True
        ).decode(encoding="utf-8")
        outcome = re.findall(pattern, query, re.M)
        for service in outcome:
            ip = re.search(ip_pattern, service)
            if ip:
                ips.append(ip.group(0))
        logger.info("K8s service ips: %s", ips)
        time.sleep(timeout)
    return ips


def find_pod_ips(num_expected, retries=3, timeout=10):
    ips = []
    attempts = 0
    while len(ips) < num_expected and attempts < retries:
        ips = []
        query = str(
            subprocess.check_output([r"kubectl get pods -o wide"], shell=True).decode(
                encoding="utf-8"
            )
        )
        pattern = r"^daliuge-daemon.*"
        ip_pattern = r"\d+\.\d+\.\d+\.\d+"
        outcome = re.findall(pattern, query, re.M)
        for pod in outcome:
            ip = re.search(ip_pattern, pod)
            if ip:
                ips.append(ip.group(0))
        logger.info("K8s pod ips: %s", ips)
        time.sleep(timeout)
    return ips


def _status_all_running(statuses):
    if statuses == []:
        return False
    for status in statuses:
        if status != "Running":
            return False
    return True


def wait_for_pods(num_expected, retries=18, timeout=10):
    all_running = False
    attempts = 0
    while not all_running and attempts < retries:
        query = str(
            subprocess.check_output([r"kubectl get pods -o wide"], shell=True).decode(
                encoding="utf-8"
            )
        )
        logger.info(query)
        pattern = r"^daliuge-daemon.*"
        outcome = re.findall(pattern, query, re.M)
        if len(outcome) < num_expected:
            all_running = False
            continue
        all_running = True
        for pod in outcome:
            if "Running" not in pod:
                all_running = False
        attempts += 1
        time.sleep(timeout)
    return all_running



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
Module with utility methods to create runtime file paths for the DALiuGE Engine.

The Engine supports implicit filepath naming if FileDROPs have no user-defined name in
the Logical Graph Template. This helps when multiple FileDROPs are created in a
construct (e.g. Scatter or Loop) and we don't run the risk of overwriting them.

This module provides the methods that support different filepath naming heuristics to
facilitate implicit filepath naming at runtime.
"""
import datetime
import logging
import re
import os
import uuid

NON_FILENAME_CHARACTERS = re.compile(fr":|{os.sep}")

def default_map():
    """
    Get the default map for the FSTRING replacement keywords
    """
    return {
        "dlg": "DALIGUE",
        "datetime": datetime.date.today().strftime("%Y-%m-%d"),
        "uid": str(uuid.uuid4()),
    } 

def base_uid_filename(uid: str, humanKey: str):
    """
    This a basic filename generator, using the UID and humandReadableKey. The function
    returns only the name of the file, and expects the full filepath to be handled by
    the function caller.

    Parameters
    ----------
    uid: str, the UID of the DROP
    humanKey: str, a human-readable key (typically a truncated UID) for user-facing
        display

    Notes
    -----
    An example DROP UID would be:

        '2024-10-30T12:01:57_0140555b-8c23-4d6a-9e24-e16c15555e8c_0'

    This approach will lead to an abridged UID along with the humanKey appended:

       '2024-10-30T12:01:57_0140555b_0/1/1'

    In this case, 0/1/1 is the "humanKey", derived from the DALiuGE Translator
    constructs (For more information on humanKey generation, refer to
    dlg.common.truncateUidToKey).

    The final part of the method replaces characters that cannot be in filenames from
    the UIDs (e.g. "/").

    Returns
    -------
        str: Base filename
    """
    if not uid:
        return None
    if not humanKey:
        humanKey = ""
    fn = str(uid).split("_")[0] + "_" + str(humanKey)
    return NON_FILENAME_CHARACTERS.sub("_", fn)

def find_dlg_fstrings(filename: str) -> list[str]:
    """
    Find f-string style braces intended for replacement by the Engine and return a list
    of the placeholder strings.

    Supports escaping the braces by using two in a row (e.g. {{example}}).

    :param filename: filename with possible f-string brackets for keyword replacement

    .. note::
        The following cases are supported by this method:
        1. No braces:
            "unique_file.txt"  # []

        2. Standard {} replacement :
            "gather_{uid}.py" -> return ['uid']

        3. Escaped braces:
            "gather_{{uid}}.py"  -> []

        4. Multiple braces
            "prefix_{uid}_{datetime}.dat"  # ['uid', 'datetime']

    :return: list of placeholder strings (incl. {})
    """
    opts = []
    try:
        opts.extend(re.findall(r'\{([^{}]+)\}', filename))
        return opts
    except TypeError as e:
        logging.warning("Data not in expected format %s",e)
        return opts


def filepath_from_string(filename: str, dirname: str = "", **kwargs) -> str:
    """
    Attempts to construct a filename from filename and possible mappings, which are
    built from a combination of FSTRING_MAP and **kwargs.

    Returns
    -------
    filename
    """

    opts = []
    fstring_map = default_map() 
    fstring_map.update(kwargs)
    if not filename and "humanKey" in fstring_map:
        return base_uid_filename(fstring_map["uid"], fstring_map["humanKey"])
    elif not filename:
        return filename

    opts.extend(find_dlg_fstrings(filename))
    for fp in opts:
        filename = filename.replace(f"{{{fp}}}", fstring_map[fp])

    return f"{dirname}/{filename}" if dirname else filename



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Contains a slurm client which generates slurm scripts from daliuge graphs.
"""

import datetime
import sys
import os
import subprocess
import shutil
import tempfile
import string
import dlg.remote as dlg_remote
from dlg.runtime import __git_version__ as git_commit

from dlg.deploy.configs import ConfigFactory, init_tpl, dlg_exec_str
from dlg.deploy.configs import DEFAULT_MON_PORT, DEFAULT_MON_HOST
from dlg.deploy.deployment_utils import find_numislands, label_job_dur
from paramiko.ssh_exception import SSHException


class SlurmClient:
    """
    parameters we can control:

    1. user group / account name (Required)
    2. whether to submit a graph, and if so provide graph path
    3. # of nodes (of Drop Managers)
    4. how long to run
    5. whether to produce offline graph vis
    6. whether to attach proxy for remote monitoring, and if so provide
        DLG_MON_HOST
        DLG_MON_PORT
    7. Root directory of the Log files (Required)
    """

    def __init__(
        self,
        dlg_root=None,
        host=None,
        acc=None,
        physical_graph_template_file=None,  # filename of physical graph template
        logical_graph=None,
        job_dur=30,
        num_nodes=None,
        run_proxy=False,
        mon_host=DEFAULT_MON_HOST,
        mon_port=DEFAULT_MON_PORT,
        logv=1,
        facility=None,
        zerorun=False,
        max_threads=0,
        sleepncopy=False,
        num_islands=None,
        all_nics=False,
        check_with_session=False,
        submit=False,
        remote=True,
        pip_name=None,
        username=None,
        ssh_key="",
        config=None,
        slurm_template=None,
        suffix=None
    ):
        ## Here, we want to separate out the following
        ## Config derived from CONFIG Factory - we replace with ini file
        ## Config derived from CLI, intended for replacement in the SLURM job script
        ##    - We want to replace these directives with the SLURM template
        ## Config derived from CLI that is used in the final script call
        ## Any leftover config - we keep as normal

        if config:
            # Do the config from the config file
            self.host = config['login_node']
            self._acc = config['account'] # superceded by slurm_template if present
            self.dlg_root = config['dlg_root']
            self.modules = config['modules']
            self.venv = config['venv'] # superceded by slurm_template if present
            self.exec_prefix = config["exec_prefix"]
            self.username = config['user'] if 'user' in config else username
            if not self.username:
                print("Username not configured in INI file, using local username...")
        else:
            # Setup SLURM environment variables using config
            config = ConfigFactory.create_config(facility=facility, user=username)
            self.host = config.getpar("host") if host is None else host
            self._acc = config.getpar("account") if (acc is None) else acc
            # self._user = config.getpar("user") if (username is None) else username

            # environment & sbatch
            self.dlg_root = config.getpar("dlg_root") if not dlg_root else dlg_root
            self.modules = config.getpar("modules")
            self.venv = config.getpar("venv")
            self.exec_prefix = config.getpar("exec_prefix")
            self.username = username
        # sbatch 
        if slurm_template:
            self._slurm_template = slurm_template
            self._num_nodes = 1 # placeholder
            self._job_dur = 1 # placeholder
        else:
            self._slurm_template = None
            if num_nodes is None:
                self._num_nodes = 1
            else:
                self._num_nodes = num_nodes
            self._job_dur = job_dur

        # self._log_root = (
        #     self._config.getpar("log_root") if (log_root is None) else log_root
        # )
        # 
        # start_dlg_cluster arguments
        self.visualise_graph = False
        self._logical_graph = logical_graph
        self._physical_graph_template_file = physical_graph_template_file
        self._run_proxy = run_proxy
        self._mon_host = mon_host
        self._mon_port = mon_port
        self._pip_name = pip_name
        self._logv = logv
        self._zerorun = zerorun
        self._max_threads = max_threads
        self._sleepncopy = sleepncopy
        if num_islands is None:
            self._num_islands = 1
        else:
            self._num_islands = num_islands
        self._all_nics = all_nics
        self._check_with_session = check_with_session
        self._submit = submit
        self._suffix = self.create_session_suffix(suffix)
        ni, nn, self._pip_name = find_numislands(self._physical_graph_template_file)
        if isinstance(ni, int) and ni >= self._num_islands:
            self._num_islands = ni
        if nn and nn >= self._num_nodes:
            self._num_nodes = nn

        # used for remote login/directory management.
        self._remote = remote
        self.ssh_key = ssh_key 
    
    def create_session_suffix(self, suffix=None):
        """
        Create a suffix to identify the session. If no suffix is specified, use the 
        current datetime setting. 

        :param: suffix, used to specify a non-datetime suffix. 
        :return: the final suffix 
        """
        if not suffix:
            return datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
        else:
            return suffix

    def get_session_dirname(self):
        """
        (pipeline name_)[Nnum_of_daliuge_nodes]_[time_stamp]
        """
        graph_name = self._pip_name.split("_")[0]  # use only the part of the graph name
        graph_name = graph_name.rsplit(".pgt.graph")[0]
        return "{0}_{1}".format(graph_name, self._suffix)
    

    def apply_slurm_template(self, template_str, session_id, dlg_root):
        """
        Given a string from a template file, use a string.Template object to perform
        safe substution on the string and replace $VALUES with the correct value 
        specified. 
        """
        intermed_slurm = string.Template(template_str) 
        ims = intermed_slurm.safe_substitute(session_id=session_id, dlg_root=dlg_root)
        print("Creating job description")
        return ims + "\n\n" + dlg_exec_str

    def create_paramater_mapping(self, session_dir, physical_graph_file):
        """
        Map the runtime or configured parameters to the session environment and SLURM 
        script paramteres, in anticipation of using substition. 
        """ 
        pardict = {}
        pardict["SESSION_ID"] = os.path.split(session_dir)[-1]
        pardict["MODULES"] = self.modules
        pardict["DLG_ROOT"] = self.dlg_root
        pardict["EXEC_PREFIX"] = self.exec_prefix
        pardict["NUM_NODES"] = str(self._num_nodes)
        pardict["JOB_DURATION"] = label_job_dur(self._job_dur)

        pardict["VENV"] = self.venv
        pardict["PIP_NAME"] = self._pip_name
        pardict["ACCOUNT"] = self._acc
        pardict["PY_BIN"] = "python3" if pardict["VENV"] else sys.executable
        pardict["LOG_DIR"] = session_dir
        pardict["GRAPH_PAR"] = (
            '--logical-graph "{0}"'.format(self._logical_graph)
            if self._logical_graph
            else (
                '--physical-graph "{0}"'.format(physical_graph_file)
                if physical_graph_file
                else ""
            )
        )
        pardict["PROXY_PAR"] = (
            "--monitor_host %s --monitor_port %d" % (self._mon_host, self._mon_port)
            if self._run_proxy
            else ""
        )
        pardict["GRAPH_VIS_PAR"] = "--dump" if self.visualise_graph else ""
        pardict["LOGV_PAR"] = "--verbose-level %d" % self._logv
        pardict["ZERORUN_PAR"] = "--zerorun" if self._zerorun else ""
        pardict["MAXTHREADS_PAR"] = "--max-threads %d" % self._max_threads
        pardict["SNC_PAR"] = "--app 1" if self._sleepncopy else "--app 0"
        pardict["NUM_ISLANDS_PAR"] = "--num_islands %d" % self._num_islands
        pardict["ALL_NICS"] = "--all_nics" if self._all_nics else ""
        pardict["CHECK_WITH_SESSION"] = (
            "--check_with_session" if self._check_with_session else ""
        )
        return pardict

    def create_job_desc(self, physical_graph_file):
        """
        Creates the slurm script from a physical graph

        This uses string.Template to apply substitutions that are linked to the 
        parameters defined at runtime. These parameters map to $VALUEs in a pre-defined
        execution command that contains the necessary parameters to run DALiuGE through
        SLURM. 
        """

        session_dir = "{0}/workspace/{1}".format(
            self.dlg_root, self.get_session_dirname()
        )
        pardict = self.create_paramater_mapping(session_dir, physical_graph_file)

        if self._slurm_template:
            slurm_str = self.apply_slurm_template(self._slurm_template, 
                                                  pardict['SESSION_ID'],
                                                  pardict['DLG_ROOT'])
            return string.Template(slurm_str).safe_substitute(pardict)

        return init_tpl.safe_substitute(pardict)

    @property
    def session_dir(self):
        return "{0}/workspace/{1}".format(
            self.dlg_root, self.get_session_dirname()
        )

    def mk_session_dir(self, dlg_root: str = ""):
        """
        Create the session directory. If dlg_root is provided it is used,
        else env var DLG_ROOT is used.
        """

        if dlg_root:  # has always preference
            self.dlg_root = dlg_root
        if self._remote and not self.dlg_root:
            print("Deploying on a remote cluster requires specifying DLG_ROOT!")
            print("Unable to create session directory!")
            return ""
        elif not self._remote:
            # locally fallback to env var
            if os.environ["DLG_ROOT"]:
                self.dlg_root = os.environ["DLG_ROOT"]
            else:
                self.dlg_root = f"{os.environ['HOME']}.dlg"
        session_dir =  self.session_dir
        if not self._remote and not os.path.exists(session_dir):
            os.makedirs(session_dir)
        if self._remote:
            command = f"mkdir -p {session_dir}"
            print(
                f"Creating remote session directory on {self.username}@{self.host}: {command}"
            )
            try:
                dlg_remote.execRemote(
                    self.host, command, username=self.username, pkeyPath=self.ssh_key
                )
            except (TypeError, SSHException) as e:
                print(
                    f"ERROR: Unable to create {session_dir} on {self.username}@{self.host}, {str(e)}"
                )
                return None

        return session_dir

    def submit_job(self):
        """
        Submits the slurm script to the requested facility

        :returns: jobId, the id of the SLURM job create on the facility. 
                  None if a remote directory could not be created or if an error occurs
                  during connection. 
        """
        jobId = None
        session_dir = self.mk_session_dir()
        if not session_dir:
            print("No session_dir created.")
            return jobId

        physical_graph_file_name = "{0}/{1}".format(session_dir, self._pip_name)
        if self._physical_graph_template_file:
            if self._remote:
                print(f"Copying PGT to: {physical_graph_file_name}")
                dlg_remote.copyTo(
                    self.host,
                    self._physical_graph_template_file,
                    physical_graph_file_name,
                    username=self.username,
                    pkeyPath=self.ssh_key,
                )
            else:
                shutil.copyfile(
                    self._physical_graph_template_file, physical_graph_file_name
                )

        job_file_name = "{0}/jobsub.sh".format(session_dir)
        job_desc = self.create_job_desc(physical_graph_file_name)

        if self._remote:
            print(f"Creating SLURM script remotely: {job_file_name}")
            tjob = tempfile.mktemp()
            with open(tjob, "w+t") as t:
                t.write(job_desc)
            dlg_remote.copyTo(
                self.host,
                tjob,
                job_file_name,
                username=self.username,
                pkeyPath=self.ssh_key,
            )
            os.remove(tjob)
        else:
            with open(job_file_name, "w") as job_file:
                job_file.write(job_desc)
            with open(os.path.join(session_dir, "git_commit.txt"), "w") as git_file:
                git_file.write(git_commit)
        if self._submit:
            if not self._remote:
                os.chdir(session_dir)  # so that slurm logs will be dumped here
                print(subprocess.check_output(["sbatch", job_file_name]))
            else:
                command = f"cd {session_dir} && sbatch --parsable {job_file_name}"
                print(f"Submitting sbatch job: {command}")
                stdout, stderr, exitStatus = dlg_remote.execRemote(
                    self.host, command, username=self.username, pkeyPath=self.ssh_key
                )
                if exitStatus != 0:
                    print(
                        f"Job submission unsuccessful: {exitStatus}, {stderr.decode()}"
                    )
                else:
                    jobId = stdout.decode()
                    print(f"Job with ID {jobId} submitted successfully.")
        else:
            print(f"Created job submission script {job_file_name}")
        return jobId



import logging
import time
import pytest

from pathlib import Path


from dlg.runtime import version  # Imported to setup DlgLogger
from dlg.apps.app_base import BarrierAppDROP
from dlg.droputils import DROPWaiterCtx
from dlg.manager.session import Session, generateLogFileName

from test.dlg_engine_testconstants import DEFAULT_TEST_REPRO, DEFAULT_TEST_GRAPH_REPRO


class MockThrowingDrop(BarrierAppDROP):
    def run(self):
        raise RuntimeError("App drop thrown")


def add_test_reprodata(graph: list):
    for drop in graph:
        drop["reprodata"] = DEFAULT_TEST_REPRO.copy()
    graph.append(DEFAULT_TEST_GRAPH_REPRO.copy())
    return graph


def test_logs(caplog):
    """
    Confirm that when we run a session in which the AppDrop experiences a runtime error,
    we produce the trackback for that application in the file.

    This acts as a regression test to make sure changes in the future don't lead to
    app/data drop tracking no longer adding appropriate attributes to the LogRecords such
    that they pass the filter setup in the Session constructor. For further information,
    review the runtime/__init__.py file.

    The test uses the pytest.caplog fixture to first confirm that:
    1. An exception is logged in the DlgLogger, and
    2. That the exception passes through the filters setup in the Session constructor, and
        properly described in the session log file.

    This aims to detect regressions when re-organising class structures or logging in the
    future.
    """

    with caplog.at_level(logging.INFO):
        with Session("log-session") as s:
            s.addGraphSpec(
                add_test_reprodata(
                    [
                        {
                            "oid": "A",
                            "categoryType": "Data",
                            "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                            "consumers": ["B"],
                        },
                        {
                            "oid": "B",
                            "categoryType": "Application",
                            "dropclass": "test.test_logs.MockThrowingDrop",
                            "sleep_time": 2,
                        },
                        {
                            "oid": "C",
                            "categoryType": "Data",
                            "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                            "producers": ["B"],
                        },
                    ]
                )
            )

            s.deploy()
            with DROPWaiterCtx(None, s.drops["C"], 300):
                s.drops["A"].write(b"x")
                s.drops["A"].setCompleted()

            # Logger needs time to get messages.
            time.sleep(5)
            logfile = Path(generateLogFileName(s._sessionDir, s.sessionId))
            exception_logged = False
            for record in caplog.records:
                if (
                    record.name == "dlg.dlg.apps.app_base"
                    and record.levelname == "ERROR"
                ):
                    exception_logged = True
                    with logfile.open("r") as f:
                        buffer = f.read()
                        assert record.name in buffer
                        assert "Traceback" in buffer
                        assert "App drop thrown" in buffer
            assert exception_logged
            logfile.unlink(missing_ok=True)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module containing command-line entry points to launch Data Manager instances
like DMs and DIMs.
"""

import logging
import os
import signal
import subprocess
import sys
import time
import re

import daemon
from lockfile.pidlockfile import PIDLockFile
from multiprocessing import Process

from .composite_manager import DataIslandManager, MasterManager
from dlg.constants import (
    NODE_DEFAULT_REST_PORT,
    ISLAND_DEFAULT_REST_PORT,
    MASTER_DEFAULT_REST_PORT,
    REPLAY_DEFAULT_REST_PORT,
    NODE_DEFAULT_RPC_PORT,
    NODE_DEFAULT_EVENTS_PORT,
)
from .node_manager import NodeManager
from .replay import ReplayManager, ReplayManagerServer
from .rest import (
    NMRestServer,
    CompositeManagerRestServer,
    MasterManagerRestServer,
)
from dlg import utils
from ..runtime import version


_terminating = False
MAX_WATCHDOG_RESTART = 10

class DlgFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        arg_pattern = re.compile(r"%\((\w+)\)")
        arg_names = [x.group(1) for x in arg_pattern.finditer(self._fmt)]
        for field in arg_names:
            if field not in record.__dict__:
                record.__dict__[field] = None

        return super().format(record)


def run_server(server, host, port):
    """
    Run the server using the options. Non-nested to avoid issues if needing to be pickled.
    :param opts: Command line options
    :return: None
    """
    server.start(host, port)


def launchServer(opts):
    # we might be called via __main__, but we want a nice logger name
    logger = logging.getLogger(f"dlg.{__name__}")
    dmName = opts.dmType.__name__

    logger.info("DALiuGE version %s running at %s", version.full_version, os.getcwd())
    logger.info("Creating %s", dmName)
    try:
        dm = opts.dmType(*opts.dmArgs, **opts.dmKwargs)
    except Exception as e:
        logger.exception(
            "Error while creating/starting our %s, exiting in shame :-(",
            dmName,
        )
        raise RuntimeError from e

    server = opts.restType(dm, opts.maxreqsize)

    # Signal handling
    def handle_signal(signNo, stack_frame):
        global _terminating # pylint: disable=global-statement
        if _terminating:
            return
        _terminating = True
        logger.info("Exiting from %s after receiving signal %s in frame %s",
                    dmName, signNo, stack_frame)

        server.stop_manager()

    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGTERM, handle_signal)
    # signal.signal(signal.SIGSEGV, handle_signal)

    if opts.watchdog_enabled:
        start_watchdog(server, opts, logger)
    else:
        run_server(server, opts.host, opts.port)


def start_watchdog(server, opts, logger):
    """
    Run the server in a separate process to keep it under observation.

    This is an experimental feature intended to secure complete runtime shutdown when we
    run potentially memory-unsafe code in child threads of the server.

    If we detect a SIGSEGV, we warn the user and attempt to re-run the server on a new
    port so that the NodeManager doesn't completely disappear and we avoid any issues with
    phantom threads using the old port.

    If we recieve any other exitcodes, we do not attempt a restart and instead following
    existing exit protocols.

    This method only acts as the watchdog, and does not support Session management or
    any explicit reconnect, and instead leaves that up to the respective manager
    implementations.

    :param server: server object that we are watching
    :param opts: runtime options
    :param logger: runtime logger
    """

    watchdog = True
    wait_period = 30
    current_restart = 0
    while watchdog:
        p = Process(target=run_server, args=(server, opts.host, opts.port))
        p.start()
        p.join()
        if p.exitcode == signal.SIGSEGV or p.exitcode == -signal.SIGSEGV:
            logger.warning(
                "Threaded server crashed with SIGSEGV (signal 11)."
                "Restarting in %s seconds...",
                wait_period
            )
            opts.port += 27 # Avoid collisions with more obvious ports, e.g. 8080 or 8888
            if current_restart < MAX_WATCHDOG_RESTART:
                current_restart +=1
            else:
                watchdog=False
                logger.warning(
                    "Watchdog service reached maximum restarts. Shutting Down..."
            )
        else:
            watchdog = False


def addCommonOptions(parser, defaultPort):
    parser.add_option(
        "-H",
        "--host",
        action="store",
        type="string",
        dest="host",
        help="The host to bind this instance on",
        default="localhost",
    )
    parser.add_option(
        "-P",
        "--port",
        action="store",
        type="int",
        dest="port",
        help="The port to bind this instance on",
        default=defaultPort,
    )
    parser.add_option(
        "-m",
        "--max-request-size",
        action="store",
        type="int",
        dest="maxreqsize",
        help="The maximum allowed HTTP request size, in MB",
        default=10,
    )
    parser.add_option(
        "-d",
        "--daemon",
        action="store_true",
        dest="daemon",
        help="Run as daemon",
        default=False,
    )
    parser.add_option(
        "-s",
        "--stop",
        action="store_true",
        dest="stop",
        help="Stop an instance running as daemon",
        default=False,
    )
    parser.add_option(
        "--status",
        action="store_true",
        dest="status",
        help="Checks if there is daemon process actively running",
        default=False,
    )
    parser.add_option(
        "-T",
        "--timeout",
        action="store",
        dest="timeout",
        type="float",
        help="Timeout used when checking for the daemon process",
        default=10,
    )
    parser.add_option(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        help="Become more verbose. The more flags, the more verbose",
    )
    parser.add_option(
        "-q",
        "--quiet",
        action="count",
        dest="quiet",
        help="Be less verbose. The more flags, the quieter",
    )
    parser.add_option(
        "-l",
        "--log-dir",
        action="store",
        type="string",
        dest="logdir",
        help="The directory where the logging files will be stored",
        default=utils.getDlgLogsDir(),
    )
    parser.add_option(
        "--watchdog",
        action="store_true",
        dest="watchdog_enabled",
        help="Enable watchdog process wrapper for server (WARNING: Experimental feature)"
    )

    parser.add_option(
        "--local-time",
        action="store_true",
        help="Use local system time when logging",
        default=False,
    )


def commonOptionsCheck(options, parser):
    # These are all exclusive
    if options.daemon and options.stop:
        parser.error("-d and -s cannot be specified together")
    if options.daemon and options.status:
        parser.error("-d and --status cannot be specified together")
    if options.stop and options.status:
        parser.error("-s and --status cannot be specified together")
    # -v and -q are exclusive
    if options.verbose and options.quiet:
        parser.error("-v and -q cannot be specified together")


def start(options, parser):
    # Perform common option checks
    commonOptionsCheck(options, parser)

    # Setup the loggers
    fileHandler = setupLogging(options)

    # Start daemon?
    if options.daemon:
        # Make sure the PID file will be created without problems
        pidDir = utils.getDlgPidDir()
        utils.createDirIfMissing(pidDir)
        pidfile = os.path.join(pidDir, "dlg%s.pid" % (options.dmAcronym))

        with daemon.DaemonContext(
            pidfile=PIDLockFile(pidfile, 1),
            files_preserve=[fileHandler.stream],
            working_directory=utils.getDlgWorkDir(),
        ):
            launchServer(options)

    # Stop daemon?
    elif options.stop:
        pidDir = utils.getDlgPidDir()
        pidfile = os.path.join(pidDir, "dlg%s.pid" % (options.dmAcronym))
        pid = PIDLockFile(pidfile).read_pid()
        if pid is None:
            sys.stderr.write("Cannot read PID file, is there an instance running?\n")
        else:
            utils.terminate_or_kill(utils.ExistingProcess(pid), 5)
            if os.path.exists(pidfile):
                sys.stderr.write(
                    "Process %d does not exist, removing PID file\n" % (pid,)
                )
                os.unlink(pidfile)

    # Check status
    elif options.status:
        socket_is_listening = utils.check_port(
            options.host, options.port, options.timeout
        )
        sys.exit(socket_is_listening is False)

    # Start directly
    else:
        working_dir = utils.getDlgWorkDir()
        tree = "/settings"
        utils.createDirIfMissing(working_dir + tree)
        os.chdir(working_dir)
        launchServer(options)


def setupLogging(opts):
    if logging.root.handlers:
        # Mmmm, somebody already did some logging, it shouldn't have been us
        # Let's reset the root handlers
        for h in logging.root.handlers[:]:
            logging.root.removeHandler(h)

    levels = [
        logging.NOTSET,
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]

    # Default is WARNING
    lidx = 3
    if opts.verbose:
        lidx -= min((opts.verbose, 3))
    elif opts.quiet:
        lidx += min((opts.quiet, 2))
    level = levels[lidx]

    # Output to files/stdout uses a command format, which can or not contain
    # optionally a session_id and drop_uid to indicate what is currently being
    # executed. This only applies to the NodeManager logs though, for which a
    # 'no_log_ids' option exists (but can be set to True).
    # We also skip logging IDs when stopping a daemon, as the infrastructure
    # won't have been set
    log_ids = opts.dmType == NodeManager and not opts.no_log_ids and not opts.stop
    fmt = "%(asctime)-15s [%(levelname)5.5s] [%(threadName)15.15s] "
    if log_ids:
        fmt += "[%(session_id)10.10s] [%(drop_uid)10.10s] "
    fmt += "%(name)s#%(funcName)s:%(lineno)s %(message)s"
    fmt = DlgFormatter(fmt)
    fmt.converter = time.localtime if opts.local_time else time.gmtime
    time_fmt =  "Local" if opts.local_time else "GMT"
    # Let's configure logging now
    # Daemons don't output stuff to the stdout
    is_daemon_argument = opts.daemon or opts.stop
    if not is_daemon_argument:
        streamHdlr = logging.StreamHandler(sys.stdout)
        streamHdlr.setFormatter(fmt)
        logging.root.addHandler(streamHdlr)

    # This is the logfile we'll use from now on
    logdir = opts.logdir
    utils.createDirIfMissing(logdir)
    logfile = os.path.join(logdir, "dlg%s.log" % (opts.dmAcronym))
    fileHandler = logging.FileHandler(logfile)
    fileHandler.setFormatter(fmt)
    logging.root.addHandler(fileHandler)

    # Per-package/module specific levels
    logging.root.setLevel(level)
    logging.getLogger("dlg").setLevel(level)
    logging.getLogger("zerorpc").setLevel(logging.WARN)

    # Assuming we have selected the default, info-level messages will not show to the
    # user. A Warning message here let's the user know something is happening without
    # us needing to modify the default logging level.
    logging.warning("Starting with level: %s...", logging.getLevelName(level))
    logging.warning("Using %s Time for logging...", time_fmt)

    return fileHandler


def dlgNM(parser, args):
    """
    Entry point for the dlg nm command
    """

    # Parse command-line and check options
    addCommonOptions(parser, NODE_DEFAULT_REST_PORT)
    parser.add_option(
        "-I",
        "--no-log-ids",
        action="store_true",
        dest="no_log_ids",
        help="Do not add associated session IDs and Drop UIDs to log statements",
        default=False,
    )
    parser.add_option(
        "--no-dlm",
        action="store_true",
        dest="noDLM",
        help="(DEPRECATED) Don't start the Data Lifecycle Manager on this NodeManager",
    )
    parser.add_option(
        "--dlm-check-period",
        type="float",
        help="Time in seconds between background DLM drop status checks (defaults to 10)",
        default=10,
    )
    parser.add_option(
        "--dlm-cleanup-period",
        type="float",
        help="Time in seconds between background DLM drop automatic cleanups (defaults to 30)",
        default=30,
    )
    parser.add_option(
        "--dlm-enable-replication",
        action="store_true",
        help="Turn on data drop automatic replication (off by default)",
    )
    parser.add_option(
        "--dlg-path",
        action="store",
        type="string",
        dest="dlgPath",
        help="Path where more DALiuGE-related libraries can be found",
        default=utils.getDlgPath(),
    )
    parser.add_option(
        "--error-listener",
        action="store",
        type="string",
        dest="errorListener",
        help="The error listener class to be used",
        default=None,
    )
    parser.add_option(
        "--event-listeners",
        action="store",
        type="string",
        dest="event_listeners",
        help="A colon-separated list of event listener classes to be used",
        default="",
    )
    parser.add_option(
        "-t",
        "--max-threads",
        action="store",
        type="int",
        dest="max_threads",
        help="Max thread pool size used for executing drops. <= 0 means use all (physical) CPUs. Default is 0.",
        default=0,
    )
    parser.add_option(
        "-p",
        "--processes",
        action="store_true",
        dest="use_processes",
        help="Use processes instead of threads to execute app drops, defaults to False",
    )
    parser.add_option(
        "--rpc_port",
        action="store",
        type="int",
        dest="rpc_port",
        help="Set the port number for the NodeManager RPC client",
        default=NODE_DEFAULT_RPC_PORT,
    )
    parser.add_option(
        "--event_port",
        action="store",
        type="int",
        dest="events_port",
        help="Set the port number for the NodeManager ZMQ client",
        default=NODE_DEFAULT_EVENTS_PORT,
    )

    (options, args) = parser.parse_args(args)

    # No logging setup at this point yet
    if options.noDLM:
        print("WARNING: --no-dlm is deprecated, use the --dlm-* options instead")
        options.dlm_check_period = 0
        options.dlm_cleanup_period = 0
        options.dlm_enable_replication = False

    # Add DM-specific options
    # Note that the host we use to expose the NodeManager itself through Pyro is
    # also used to expose the Sessions it creates
    options.dmType = NodeManager
    options.dmArgs = ()
    options.dmKwargs = {
        "dlm_check_period": options.dlm_check_period,
        "dlm_cleanup_period": options.dlm_cleanup_period,
        "dlm_enable_replication": options.dlm_enable_replication,
        "dlgPath": options.dlgPath,
        "host": options.host,
        "rpc_port": options.rpc_port,
        "events_port": options.events_port,
        "error_listener": options.errorListener,
        "event_listeners": list(filter(None, options.event_listeners.split(":"))),
        "max_threads": options.max_threads,
        "use_processes": options.use_processes,
        "logdir": options.logdir,
        "use_local_time": options.local_time,
    }
    options.dmAcronym = "NM"
    options.restType = NMRestServer

    start(options, parser)


def dlgCompositeManager(parser, args, dmType, acronym, dmPort, dmRestServer):
    """
    Common entry point for the dlgDIM and dlgMM command-line scripts. It
    starts the corresponding CompositeManager and exposes it through a
    REST interface.
    """

    # Parse command-line and check options
    addCommonOptions(parser, dmPort)
    parser.add_option(
        "-N",
        "--nodes",
        action="store",
        type="string",
        dest="nodes",
        help="Comma-separated list of node names managed by this %s" % (acronym),
        default="",
    )
    parser.add_option(
        "-k",
        "--ssh-pkey-path",
        action="store",
        type="string",
        dest="pkeyPath",
        help="Path to the private SSH key to use when connecting to the nodes",
        default=None,
    )
    parser.add_option(
        "--dmCheckTimeout",
        action="store",
        type="int",
        dest="dmCheckTimeout",
        help="Maximum timeout used when automatically checking for DM presence",
        default=10,
    )
    parser.add_option(
        "--dump_graphs",
        action="store_true",
        dest="dump_graphs",
        help="Store physical graphs submitted to the manager in the workspace directory",
        default=False,
    )
    (options, args) = parser.parse_args(args)

    # Add DIM-specific options
    options.dmType = dmType
    options.dmArgs = ([s for s in options.nodes.split(",") if s],)

    options.dmKwargs = {
        "pkeyPath": options.pkeyPath,
        "dmCheckTimeout": options.dmCheckTimeout,
        "dump_graphs": options.dump_graphs,
    }
    options.dmAcronym = acronym
    options.restType = dmRestServer

    start(options, parser)


def dlgDIM(parser, args):
    """
    Entry point for the dlg dim command
    """
    dlgCompositeManager(
        parser,
        args,
        DataIslandManager,
        "DIM",
        ISLAND_DEFAULT_REST_PORT,
        CompositeManagerRestServer,
    )


def dlgMM(parser, args):
    """
    Entry point for the dlg mm command
    """
    dlgCompositeManager(
        parser,
        args,
        MasterManager,
        "MM",
        MASTER_DEFAULT_REST_PORT,
        MasterManagerRestServer,
    )


def dlgReplay(parser, args):
    """
    Entry point for the dlg replay command
    """

    # Parse command-line and check options
    addCommonOptions(parser, REPLAY_DEFAULT_REST_PORT)
    parser.add_option(
        "-S",
        "--status-file",
        action="store",
        dest="status_file",
        help="File containing a continuous graph status dump",
        default=None,
    )
    parser.add_option(
        "-g",
        "--graph-file",
        action="store",
        type="string",
        dest="graph_file",
        help="File containing a physical graph dump",
        default=None,
    )
    (options, args) = parser.parse_args(args)

    if not options.graph_file:
        parser.error("Missing mandatory graph file")
    if not options.status_file:
        parser.error("Missing mandatory status file")

    # Add DM-specific options
    options.dmType = ReplayManager
    options.dmArgs = (options.graph_file, options.status_file)
    options.dmKwargs = {}
    options.dmAcronym = "RP"
    options.restType = ReplayManagerServer

    start(options, parser)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
from abc import abstractmethod
import random
import os
import logging
from typing import Union

from dlg.drop import AbstractDROP, track_current_drop
from dlg.data.io import (
    DataIO,
    OpenMode,
    NullIO,
)
from dlg.ddap_protocol import (
    ChecksumTypes,
    DROPStates,
)
from dlg.utils import isabs, createDirIfMissing


checksum_disabled = "DLG_DISABLE_CHECKSUM" in os.environ
try:
    from crc32c import crc32c  # @UnusedImport

    _checksumType = ChecksumTypes.CRC_32C
except ImportError:
    from binascii import crc32 # pylint: disable=unused-import

    _checksumType = ChecksumTypes.CRC_32

logger = logging.getLogger(f"dlg.{__name__}")


##
# @brief Data
# @details A generic Data drop, whose functionality can be provided by an arbitrary class, as specified in the 'dropclass' component parameter. It is not useful without additional development.
# @par EAGLE_START
# @param category Data
# @param tag template
# @param dropclass dlg.data.drops.data_base.DataDROP/String/ComponentParameter/NoPort/ReadOnly//False/False/The python class that implements this data component
# @param base_name data_base/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param persist False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component contains data that should not be deleted after execution
# @param io /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Input Output port
# @par EAGLE_END
class DataDROP(AbstractDROP):
    """
    A DataDROP is a DROP that stores data for writing with
    an AppDROP, or reading with one or more AppDROPs.

    DataDROPs have two different modes: "normal" and "streaming".
    Normal DataDROPs will wait until the COMPLETED state before being
    available as input to an AppDROP, while streaming AppDROPs may be
    read simutaneously with writing by chunking drop bytes together.

    This class contains two methods that need to be overrwritten:
    `getIO`, invoked by AppDROPs when reading or writing to a drop,
    and `dataURL`, a getter for a data URI uncluding protocol and address
    parsed by function `IOForURL`.
    """

    def incrRefCount(self):
        """
        Increments the reference count of this DROP by one atomically.
        """
        with self._refLock:
            self._refCount += 1

    def decrRefCount(self):
        """
        Decrements the reference count of this DROP by one atomically.
        """
        with self._refLock:
            self._refCount -= 1

    @track_current_drop
    def open(self, **kwargs):
        """
        Opens the DROP for reading, and returns a "DROP descriptor"
        that must be used when invoking the read() and close() methods.
        DROPs maintain a internal reference count based on the number
        of times they are opened for reading; because of that after a successful
        call to this method the corresponding close() method must eventually be
        invoked. Failing to do so will result in DROPs not expiring and
        getting deleted.
        """
        if self.status != DROPStates.COMPLETED:
            raise Exception(
                "%r is in state %s (!=COMPLETED), cannot be opened for reading"
                % (self, self.status)
            )

        io = self.getIO()
        logger.debug("Opening drop %s", self.oid)
        io.open(OpenMode.OPEN_READ, **kwargs)

        # Save the IO object in the dictionary and return its descriptor instead
        while True:
            descriptor = random.SystemRandom().randint(-(2**31), 2**31 - 1)
            if descriptor not in self._rios:
                break
        self._rios[descriptor] = io

        # This occurs only after a successful opening
        self.incrRefCount()
        self._fire("open")

        return descriptor

    @track_current_drop
    def close(self, descriptor, **kwargs):
        """
        Closes the given DROP descriptor, decreasing the DROP's
        internal reference count and releasing the underlying resources
        associated to the descriptor.
        """
        self._checkStateAndDescriptor(descriptor)

        # Decrement counter and then actually close
        self.decrRefCount()
        io = self._rios.pop(descriptor)
        io.close(**kwargs)

    def _closeWriters(self):
        """
        Close our writing IO instance.
        If written externally, self._wio will have remained None
        """
        if self._wio:
            try:
                self._wio.close()
            except IOError:
                pass  # this will make sure that a previous issue does not cause the graph to hang!
                # raise Exception("Problem closing file!")
            self._wio = None

    def read(self, descriptor, count=65536, **kwargs):
        """
        Reads `count` bytes from the given DROP `descriptor`.
        """
        self._checkStateAndDescriptor(descriptor)
        io = self._rios[descriptor]
        return io.read(count, **kwargs)

    def _checkStateAndDescriptor(self, descriptor):
        if self.status != DROPStates.COMPLETED:
            raise Exception(
                "%r is in state %s (!=COMPLETED), cannot be read" % (self, self.status)
            )
        if descriptor is None:
            raise ValueError("Illegal empty descriptor given")
        if descriptor not in self._rios:
            raise Exception(
                "Illegal descriptor %d given, remember to open() first" % (descriptor)
            )

    def isBeingRead(self):
        """
        Returns `True` if the DROP is currently being read; `False`
        otherwise
        """
        with self._refLock:
            return self._refCount > 0

    @track_current_drop
    def write(self, data: Union[bytes, memoryview]):
        """
        Writes the given `data` into this DROP. This method is only meant
        to be called while the DROP is in INITIALIZED or WRITING state;
        once the DROP is COMPLETE or beyond only reading is allowed.
        The underlying storage mechanism is responsible for implementing the
        final writing logic via the `self.writeMeta()` method.
        """

        if self.status not in [DROPStates.INITIALIZED, DROPStates.WRITING]:
            raise Exception("No more writing expected")

        if not isinstance(data, (bytes, memoryview, str)):
            raise Exception("Data type not of binary type: ", type(data).__name__)

        # We lazily initialize our writing IO instance because the data of this
        # DROP might not be written through this DROP
        if not self._wio:
            self._wio = self.getIO()
            try:
                self._wio.open(OpenMode.OPEN_WRITE)
            except IOError as e:
                self.status = DROPStates.ERROR
                raise IOError("Problem opening drop for write!") from e
        nbytes = self._wio.write(data)
        nbytes = 0 if nbytes is None else nbytes

        dataLen = len(data)
        if nbytes != dataLen:
            # TODO: Maybe this should be an actual error?
            logger.warning(
                "Not all data was correctly written by %s (%d/%d bytes written)",
                self,
                nbytes,
                dataLen,
            )

        # see __init__ for the initialization to None
        if self._size is None:
            self._size = 0
        self._size += nbytes

        # Trigger our streaming consumers
        if self._streamingConsumers:
            for streamingConsumer in self._streamingConsumers:
                streamingConsumer.dataWritten(self.uid, data)

        # Update our internal checksum
        if not checksum_disabled:
            self._updateChecksum(data)

        # If we know how much data we'll receive, keep track of it and
        # automatically switch to COMPLETED
        if self._expectedSize > 0:
            remaining = self._expectedSize - self._size
            if remaining > 0:
                self.status = DROPStates.WRITING
            else:
                if remaining < 0:
                    logger.warning(
                        "Received and wrote more bytes than expected: %d",
                        -remaining,
                    )
                logger.debug(
                    "Automatically moving %r to COMPLETED, all expected data arrived",
                    self,
                )
                self.setCompleted()
        else:
            self.status = DROPStates.WRITING

        return nbytes

    def _updateChecksum(self, chunk):
        # see __init__ for the initialization to None
        if self._checksum is None:
            self._checksum = 0
            self._checksumType = _checksumType
        if isinstance(chunk, str):
            chunk = bytes(chunk, encoding="utf8")
        self._checksum = crc32c(chunk, self._checksum)

    @property
    def checksum(self):
        """
        The checksum value for the data represented by this DROP. Its
        value is automatically calculated if the data was actually written
        through this DROP (using the `self.write()` method directly or
        indirectly). In the case that the data has been externally written, the
        checksum can be set externally after the DROP has been moved to
        COMPLETED or beyond.

        :see: `self.checksumType`
        """
        if self.status == DROPStates.COMPLETED and self._checksum is None:
            # Generate on the fly
            io = self.getIO()
            io.open(OpenMode.OPEN_READ)
            data = io.read(65536)
            if isinstance(data, str):
                data = bytes(data, encoding="utf8")
            while data is not None and len(data) > 0:
                self._updateChecksum(data)
                data = io.read(65536)
            io.close()
        return self._checksum

    @checksum.setter
    def checksum(self, value):
        if self._checksum is not None:
            raise Exception(
                "The checksum for DROP %s is already calculated, cannot overwrite with new value"
                % (self)
            )
        if self.status in [DROPStates.INITIALIZED, DROPStates.WRITING]:
            raise Exception(
                "DROP %s is still not fully written, cannot manually set a checksum yet"
                % (self)
            )
        self._checksum = value

    @property
    def checksumType(self):
        """
        The algorithm used to compute this DROP's data checksum. Its value
        if automatically set if the data was actually written through this
        DROP (using the `self.write()` method directly or indirectly). In
        the case that the data has been externally written, the checksum type
        can be set externally after the DROP has been moved to COMPLETED
        or beyond.

        :see: `self.checksum`
        """
        return self._checksumType

    @checksumType.setter
    def checksumType(self, value):
        if self._checksumType is not None:
            raise Exception(
                "The checksum type for DROP %s is already set, cannot overwrite with new value"
                % (self)
            )
        if self.status in [DROPStates.INITIALIZED, DROPStates.WRITING]:
            raise Exception(
                "DROP %s is still not fully written, cannot manually set a checksum type yet"
                % (self)
            )
        self._checksumType = value

    def _map_input_ports_to_params(self):
        """
        Map the input ports that are on the drop to the its parameters

        This method performs the following steps:

            - Iterate through each producer, finding the portname and value of the
            port associated with that producers

            - Iterate through the input port names of _this_ drop, and match the UID
            and name to the producer map, and then getting the value.

            - Finally, match the value of the named input drop with a DROP parameter (
            if it exists).

        It is expected that this method be used within the child DataDrop class that is
        inheriting this method; see the FileDrop class implemenetation for an example
        use case.
        """

        try:
            dropInputPorts = self.parameters["producers"]
        except KeyError:
            logging.debug("No producers available for drop: %s", self.uid)
            return

        producerPortValueMap = {}  # Map Producer UIDs to a portname
        finalDropPortMap = {}  # Final mapping of named port to value stored in producer

        for p in self.producers:
            producerUid = p.uid
            producerPortValueMap[producerUid] = {}
            params = p.parameters["outputs"]
            for param in params:
                try:
                    key = list(param.keys())[0]
                except AttributeError:
                    logging.debug("Producer %s does not have named ports", p.uid)
                    continue
                portName = param[key]
                portValue = ""
                if portName in p.parameters:
                    portValue = p.parameters[param[key]]
                producerPortValueMap[producerUid][portName] = portValue

        for port in dropInputPorts:
            try:
                port.items()
            except AttributeError:
                logging.debug("Producer %s does not have named ports", port.uid)
                continue
            for uid, input_port_name in port.items():
                try:
                    ouput_port_name = self.parameters["port_map"][input_port_name]
                    if ouput_port_name in producerPortValueMap[uid]:
                        finalDropPortMap[input_port_name] = producerPortValueMap[uid][
                            ouput_port_name]
                except KeyError:
                    logging.warning("%s not available.", input_port_name)

        for portname in finalDropPortMap:
            if portname in self.parameters:
                self.parameters[portname] = finalDropPortMap[portname]

    @abstractmethod
    def getIO(self) -> DataIO:
        """
        Returns an instance of one of the `dlg.io.DataIO` instances that
        handles the data contents of this DROP.
        """

    def delete(self):
        """
        Deletes the data represented by this DROP.
        """
        self.getIO().delete()

    def exists(self):
        """
        Returns `True` if the data represented by this DROP exists indeed
        in the underlying storage mechanism
        """
        return self.getIO().exists()

    @property
    @abstractmethod
    def dataURL(self) -> str:
        """
        A URL that points to the data referenced by this DROP. Different
        DROP implementations will use different URI schemes.
        """


class PathBasedDrop(object):
    """
    Base class for data drops that handle paths (i.e., file and directory drops)
    """

    _path: str = None

    def get_dir(self, dirname):
        """
        dirname will be based on the current working directory
        If we have a session, it goes into the path as well
        (most times we should have a session BTW, we should expect *not* to
        have one only during testing)

        :param dirname: str

        :returns dir
        """
        if isabs(dirname):
            return dirname

        parts = []
        if self._dlg_session_id:
            parts.append(".")
        else:
            parts.append("/tmp/daliuge_tfiles")
        if dirname:
            parts.append(dirname)

        the_dir = os.path.abspath(os.path.normpath(os.path.join(*parts)))
        logger.debug("Path used for drop: %s", the_dir)
        createDirIfMissing(the_dir)
        return the_dir

    @property
    def path(self) -> str:
        return self._path


##
# @brief NULL
# @details A Drop not storing any data (useful for just passing on events)
# @par EAGLE_START
# @param category Memory
# @param tag daliuge
# @param dropclass dlg.data.drops.data_base.NullDROP/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param base_name data_base/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param persist True/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component contains data that should not be deleted after execution
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param io /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Input Output port
# @par EAGLE_END
class NullDROP(DataDROP):
    """
    A DROP that doesn't store any data.
    """

    def getIO(self):
        return NullIO()

    @property
    def dataURL(self) -> str:
        return "null://"


class EndDROP(NullDROP):
    """
    A DROP that ends the session when reached
    """



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
A very simple implementation of an HSM, without knowing much actually about how
HSMs work and what they actually offer in terms of APIs.

@author: rtobar
"""

import logging

from ..hsm import store

logger = logging.getLogger(f"dlg.{__name__}")


class HierarchicalStorageManager(object):
    def __init__(self):
        self._stores = []
        self.addStore(store.MemoryStore())
        self.addStore(store.FileSystemStore("/", "/tmp/daliuge_tfiles"))

    def addStore(self, newStore):
        """
        @param newStore store.AbstractStore
        """
        logger.debug("Adding store to HSM: %s", str(newStore))
        self._stores.append(newStore)

    def getSlowestStore(self):
        """
        :return store.AbstractStore:
        """
        return self._stores[-1]



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2020
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

from dlg.common import tool


def include_dir(_parser, _args):
    from . import get_include_dir

    print(get_include_dir())


def register_commands():
    manager_group = tool.CommandGroup("managers", "DROP Manager Commands")
    tool.cmdwrap("nm", manager_group, "Starts a Node Manager", "dlg.manager.cmdline:dlgNM")
    tool.cmdwrap("dim", manager_group, "Starts a Drop Island Manager", "dlg.manager.cmdline:dlgDIM")
    tool.cmdwrap("mm", manager_group, "Starts a Master Manager", "dlg.manager.cmdline:dlgMM")
    tool.cmdwrap("replay", manager_group, "Starts a Replay Manager", "dlg.manager.cmdline:dlgReplay")
    tool.cmdwrap(
        "daemon", manager_group,
        "Starts a DALiuGE Daemon process",
        "dlg.manager.proc_daemon:run_with_cmdline",
    )
    tool.cmdwrap(
        "proxy", manager_group,
        "A reverse proxy to be used in restricted environments to contact the Drop Managers",
        "dlg.deploy.dlg_proxy:run",
    )
    tool.cmdwrap(
        "monitor", manager_group,
        "A proxy to be used in conjunction with the dlg proxy in restricted environments",
        "dlg.deploy.dlg_monitor:run",
    )
    manager_utilities = tool.CommandGroup("utilities", "Utility commands")
    tool.cmdwrap(
        "include_dir", manager_utilities,
        "Print the directory where C header files can be found",
        include_dir,
    )

    remote_group = tool.CommandGroup("remote", "Remote environment configuration and "
                                               "deployment")
    tool.cmdwrap(
        "remote-submit",remote_group,
        "Create a DALiuGE graph to a remote computing environment",
        "dlg.deploy.create_dlg_job:run", 
    )
    tool.cmdwrap("config", remote_group,
                 "Manage dlg config environment",
                 "dlg.deploy.configs.config_manager:run")




#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""Utilities to emulate the `dask.delayed` function"""

from __future__ import annotations

import contextlib
import dill
import base64
import logging
import pickle
import socket
import struct
import time

from inspect import signature

from dlg import utils, droputils
from dlg.common import dropdict
from dlg.data.drops.memory import parse_pydata
from dlg.ddap_protocol import DROPStates
from dlg.apps.app_base import BarrierAppDROP
from dlg.exceptions import InvalidDropException

logger = logging.getLogger(f"dlg.{__name__}")


class ResultTransmitter(BarrierAppDROP):
    """Collects data from all inputs and transmits it to whomever connects to
    the given host/port"""

    def initialize(self, **kwargs):
        BarrierAppDROP.initialize(self, input_error_threshold=100, **kwargs)
        self.host = self._popArg(kwargs, "host", "localhost")
        self.port = self._popArg(kwargs, "port", None)
        if self.port is None:
            raise InvalidDropException(self, "Missing port parameter")

    def run(self):
        def read_result(x):
            if x.status == DROPStates.ERROR:
                return "Error"
            try:
                content = pickle.loads(droputils.allDropContents(x))
            except EOFError:
                content = None
            return content

        results = map(read_result, self.inputs)  # @UndefinedVariable
        results = list(results)
        if len(self.inputs) == 1:
            results = results[0]
        results = pickle.dumps(results)

        s = socket.socket()
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        logger.debug(
            "Attempting to bind ResultTransmitter to %s:%i", self.host, self.port
        )
        s.bind((self.host, self.port))
        s.listen(1)
        client, _ = s.accept()
        with contextlib.closing(client):
            client = client.makefile("wb")
            client.write(struct.pack(">i", len(results)))
            client.write(results)


def _get_client(**kwargs):
    if "client" in kwargs:
        return kwargs["client"]

    from .manager.client import NodeManagerClient
    from dlg import constants

    host = kwargs.get("host", "localhost")
    port = kwargs.get("port", constants.NODE_DEFAULT_REST_PORT)
    timeout = kwargs.get("timeout", None)
    return NodeManagerClient(host, port, timeout)


def _is_list_of_delayeds(x):
    return isinstance(x, (list, tuple)) and len(x) > 0 and isinstance(x[0], _DataDrop)


def compute(value, **kwargs):
    """Returns the result of the (possibly) delayed computation by sending
    the graph to a Drop Manager and waiting for the result to arrive back"""

    # Support calling compute with a list of DelayedDrops
    if _is_list_of_delayeds(value):
        value = _DelayedDrops(*value)

    graph = value.get_graph()
    port = 10001
    # Add one final application that will wait for all results
    # and transmit them back to us
    transmitter_oid = "-1"
    transmitter = dropdict(
        {
            "categoryType": "Application",
            #            "categoryType": CategoryType.APPLICATION,
            # "Application": "dlg.dask_emulation.ResultTransmitter",
            "dropclass": "dlg.dask_emulation.ResultTransmitter",
            "oid": transmitter_oid,
            "uid": transmitter_oid,
            "port": port,
            "name": "result transmitter",
        }
    )
    for leaf_oid in droputils.get_leaves(graph.values()):
        graph[leaf_oid].addConsumer(transmitter)
    graph[transmitter_oid] = transmitter

    graph = list(graph.values())

    # Submit and wait
    session_id = "session-%f" % time.time()
    client = _get_client(**kwargs)
    client.create_session(session_id)
    client.append_graph(session_id, graph)
    client.deploy_session(session_id, completed_uids=droputils.get_roots(graph))

    timeout = kwargs.get("timeout", None)
    s = utils.connect_to("localhost", port, 10)
    s.settimeout(timeout)
    with contextlib.closing(s):
        s = s.makefile("rb")
        nbytes = struct.unpack(">i", s.read(4))[0]
        ret = pickle.loads(s.read(nbytes))
        logger.info("Received %r from graph computation", ret)
        return ret


class _DelayedDrop(object):
    _drop_count = 0

    def __init__(self, producer=None):
        self._dropdict = None
        self.producer = producer
        self.inputs = []

    @property
    def next_drop_oid(self):
        i = _DelayedDrop._drop_count
        _DelayedDrop._drop_count += 1
        return i

    @property
    def dropdict(self):
        if self._dropdict is None:
            self._dropdict = self.make_dropdict()
        return self._dropdict

    def reset(self):
        self._dropdict = None

    @property
    def oid(self):
        return self.dropdict["oid"]

    def compute(self, **kwargs):
        return compute(self, **kwargs)

    def get_graph(self):
        _DelayedDrop._drop_count = 0
        graph = {}
        visited = set()
        self.to_physical_graph(visited, graph)
        for d in visited:
            d.reset()
        return graph

    def append_to_graph(self, visited, graph):
        if self in visited:
            return
        oid = str(self.next_drop_oid)
        dd = self.dropdict
        dd["oid"] = oid
        dd["inputs"] = []

        visited.add(self)
        graph[oid] = dd
        logger.debug("Appended %r/%s to the Physical Graph", self, oid)

    def to_physical_graph(self, visited, graph):
        self.append_to_graph(visited, graph)
        dependencies = self.inputs
        if self.producer:
            dependencies.append(self.producer)
        for d in dependencies:
            if isinstance(d, list):
                d = tuple(d)
            if d in visited:
                self.add_upstream(d)
                continue

            d = d.to_physical_graph(visited, graph)
            self.add_upstream(d)

        return self

    def add_upstream(self, upstream: _DelayedDrop):
        """Link the given drop as either a producer or input of this drop"""
        self_dd = self.dropdict
        up_dd = upstream.dropdict
        if isinstance(self, _DataDrop):
            self_dd.addProducer(up_dd)
            logger.debug(
                "Set %r/%s as producer of %r/%s",
                upstream,
                upstream.oid,
                self,
                self.oid,
            )
        else:
            logger.debug(
                "Set %r/%s as input of %r/%s",
                upstream,
                upstream.oid,
                self,
                self.oid,
            )


class _Listifier(BarrierAppDROP):
    """Returns a list with all objects as contents"""

    def run(self):
        logger.debug("Listifier running...")
        self.outputs[0].write(
            pickle.dumps(
                [pickle.loads(droputils.allDropContents(x)) for x in self.inputs]
            )
        )


class _DelayedDrops(_DelayedDrop):
    """One or more _DelayedDrops treated as a single item"""

    def __init__(self, *drops):
        super(_DelayedDrops, self).__init__()
        self.drops = drops
        self.inputs.extend(drops)
        logger.debug("Created %r", self)

    def to_physical_graph(self, visited, graph):
        output = _DataDrop(producer=self)
        output.append_to_graph(visited, graph)

        self.append_to_graph(visited, graph)
        output.add_upstream(self)

        for d in self.drops:
            d.to_physical_graph(visited, graph)
            self.add_upstream(d)

        return output

    def __iter__(self):
        return iter(self.drops)

    def __len__(self):
        return len(self.drops)

    def __getitem__(self, i):
        return self.drops[i]

    def add_upstream(self, upstream: _DelayedDrop):
        _DelayedDrop.add_upstream(self, upstream)
        self.dropdict["inputs"].append(upstream.oid)

    def make_dropdict(self):
        return dropdict(
            {
                "categoryType": "Application",
                "dropclass": "dlg.dask_emulation._Listifier",
                "name": "listifier",
                "inputs": self.inputs,
            }
        )

    def __repr__(self):
        return "<_DelayedDrops n=%d>" % (len(self.drops),)


class _AppDrop(_DelayedDrop):
    """Defines a PyFuncApp drop for a given function `f`"""

    def __init__(self, f, nout):
        _DelayedDrop.__init__(self)
        self.f = f
        self.fname = None
        if hasattr(f, "__name__"):
            self.fname = f.__name__
        self.fcode, self.fdefaults = utils.serialize_func(f)
        # self.fcode = inspect.getsource(f)
        self.original_kwarg_names = []
        self.original_arg_names = []
        self.nout = nout
        logger.debug("Created %r", self)

    def make_dropdict(self):
        self.kwarg_names = list(self.original_kwarg_names)
        self.arg_names = list(self.original_arg_names)
        self.kwarg_names.reverse()
        self.arg_names.reverse()
        my_dropdict = dropdict(
            {
                # "oid": uuid.uuid1(),
                "categoryType": "Application",
                "dropclass": "dlg.apps.pyfunc.PyFuncApp",
                "func_arg_mapping": {},
                "inputs": self.inputs,
            }
        )
        if self.fname is not None:
            simple_fname = self.fname.split(".")[-1]
            my_dropdict["func_name"] = self.fname
            my_dropdict["name"] = simple_fname
        if self.fcode is not None:
            my_dropdict["func_code"] = self.fcode
            logger.debug("func_code provided: %s", dill.loads(base64.b64decode(self.fcode)))
        if self.fdefaults:
            # APPLICATION ARGUMENTS
            my_dropdict["func_defaults"] = self.fdefaults
        return my_dropdict

    def add_upstream(self, upstream):
        _DelayedDrop.add_upstream(self, upstream)
        if self.kwarg_names:
            name = self.kwarg_names.pop()
            if name is not None:
                logger.debug(
                    "Adding %s/%s to function mapping for %s",
                    name,
                    upstream.oid,
                    self.fname,
                )
                self.dropdict["inputs"].append({upstream.oid: name})
                self.dropdict["func_arg_mapping"][name] = upstream.oid
        if self.arg_names:
            name = self.arg_names.pop()
            if name is not None:
                logger.debug(
                    "Adding %s/%s to function mapping for %s",
                    name,
                    upstream.oid,
                    self.fname,
                )
                self.dropdict["inputs"].append({upstream.oid: name})
                self.dropdict["func_arg_mapping"][name] = upstream.oid

    def _to_delayed_arg(self, arg):
        logger.info("Turning into delayed arg for %r: %r", self, arg)
        if isinstance(arg, _DelayedDrop):
            return arg

        # Turn lists/tuples of _DataDrop objects into a _DelayedDrops
        if _is_list_of_delayeds(arg):
            return _DelayedDrops(*arg)

        # Plain data gets turned into a _DataDrop
        return _DataDrop(pydata=arg)

    def __call__(self, *args, **kwargs):
        logger.debug(
            "Delayed function %s called with %d args and %d kwargs",
            self.fname,
            len(args),
            len(kwargs),
        )

        for name, arg in kwargs.items():
            self.inputs.append(self._to_delayed_arg(arg))
            self.original_kwarg_names.append(name)

        leftover_args = [
            x
            for x in list(signature(self.f).parameters.keys())
            if x not in list(kwargs.keys())
        ]

        iter_range = min(len(args), len(leftover_args))
        for i in range(iter_range):
            self.inputs.append(self._to_delayed_arg(args[i]))  # , leftover_args[i]))
            self.original_arg_names.append(leftover_args[i])

        if self.nout is None:
            return _DataDrop(producer=self)

        return _DataDropSequence(nout=self.nout, producer=self)

    def __repr__(self):
        return "<_DelayedApp fname=%s, nout=%s>" % (self.fname, str(self.nout))


_no_data = object()


class _DataDrop(_DelayedDrop):
    """Defines an in-memory drop"""

    def __init__(self, producer=None, pydata=_no_data):
        _DelayedDrop.__init__(self, producer)

        if bool(producer is None) == bool(pydata is _no_data):
            raise ValueError("either producer or pydata must be not None")
        self.pydata = pydata
        logger.debug("Created %r", self)

    def make_dropdict(self):
        my_dropdict = dropdict(
            {
                # "oid": uuid.uuid1(),
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            }
        )
        if hasattr(self, "pydata"):
            my_dropdict["pydata"] = parse_pydata(self.pydata)
        return my_dropdict

    def __repr__(self):
        if not self.producer:
            return "<_DataDrop, pydata=%r>" % (self.pydata,)
        return "<_DataDrop, producer=%r>" % self.producer


def getitem(x, i):
    """
    Helper function to ensure the delayed() __getitem__ gets the appropriate arguments
    in the correct order.
    """
    try:
        return x[i]
    except TypeError:
        return x


class _DataDropSequence(_DataDrop):
    """One or more _DataDrops that can be subscribed"""

    def __init__(self, nout, producer):
        super(_DataDrop, self).__init__(producer=producer)
        self.nout = nout
        logger.debug("Created %r", self)
        self._drops = []

    def __iter__(self):
        for i in range(self.nout):
            yield self[i]

    def __len__(self):
        return self.nout

    def __getitem__(self, i):
        return delayed(getitem)(x=self, i=i)

    def __repr__(self):
        return "<_DataDropSequence nout=%d, producer=%r>" % (
            self.nout,
            self.producer,
        )


def delayed(x, *args, **kwargs):
    """Like dask.delayed, but quietly swallowing anything other than `nout`"""
    if "nout" in kwargs:
        nout = kwargs["nout"]
    elif args:
        nout = args[0]
    else:
        nout = None

    if callable(x):
        return _AppDrop(x, nout=nout)
        # return x(*args, **kwargs)
    return _DataDrop(pydata=x)
